"""Marimo adapter for `topovec.mgl` scenes and headless image containers.

This module does not define rendering semantics. Instead it mirrors the
backend-agnostic scene/property model from :mod:`topovec.mgl` into anywidget
state so the same scenes can be reused in notebooks, headless scripts, and
other frontends.
"""

from __future__ import annotations

import base64
import copy
import functools
import io
from dataclasses import dataclass
from typing import Any

import numpy as np
from PIL import Image

from ..core import ListProperty, NumericProperty, Property, System, UpdateMask
from ..mgl.camera import Camera
from ..viewer import session as _viewer_session_core
from ..viewer.residency import SceneResidencyManager
from ._assets import (
    camera_viewport_css,
    camera_viewport_esm,
    director_field_inspect_css,
    director_field_inspect_esm,
)

DEFAULT_CONTROLS_TEXT = (
    "Drag: orbit | Shift+drag: pan | Wheel: zoom | "
    "Shift+wheel: roll"
)

PROJECTION_LABELS = ("Perspective", "Orthographic")
DEFAULT_NUMERIC_STEP = 0.001


@dataclass(slots=True)
class DirectorFieldView:
    """Bundle returned by :func:`inspect_view`.

    The bundle keeps the Python-side :class:`RenderSession` together with the
    marimo widget so advanced callers can inspect or control the live session
    without reassembling the UI manually.
    """

    session: "RenderSession"
    widget: Any
    layout: Any


@dataclass(slots=True)
class _SessionMessageOutcome:
    handled: bool
    publish_frame: bool = False
    publish_camera: bool = False
    publish_properties: bool = False
    publish_summary: bool = False


def serialize_camera(camera: Camera) -> dict[str, Any]:
    """Convert a live camera into a JSON-friendly backend-neutral snapshot."""
    return {
        "look_at": [float(v) for v in np.asarray(camera.look_at, dtype=float)],
        "look_direction": [
            float(v) for v in np.asarray(camera.look_direction, dtype=float)
        ],
        "look_up": [float(v) for v in np.asarray(camera.look_up, dtype=float)],
        "look_scale": float(camera.look_scale),
        "perspective": bool(camera.perspective_is_on),
    }


def restore_camera(camera: Camera, state: dict[str, Any]) -> Camera:
    """Apply a serialized snapshot back to an existing camera instance."""
    camera.look_at = np.asarray(state["look_at"], dtype=np.float32)
    camera.look_direction = np.asarray(state["look_direction"], dtype=np.float32)
    camera.look_up = np.asarray(state["look_up"], dtype=np.float32)
    camera.look_scale = float(state["look_scale"])
    camera.perspective_is_on = bool(state["perspective"])
    return camera.normalize()


def _normalize_camera_state(camera_state: dict[str, Any]) -> dict[str, Any]:
    camera = Camera(perspective=bool(camera_state["perspective"]))
    return serialize_camera(restore_camera(camera, camera_state))


def camera_state_signature(camera_state: dict[str, Any]) -> tuple[Any, ...]:
    """Create a rounded signature used to compare camera states."""
    state = _normalize_camera_state(camera_state)

    def rounded(values: list[float]) -> tuple[float, ...]:
        return tuple(round(float(v), 6) for v in values)

    return (
        rounded(state["look_at"]),
        rounded(state["look_direction"]),
        rounded(state["look_up"]),
        round(float(state["look_scale"]), 6),
        bool(state["perspective"]),
    )


def _format_python_float(value: float) -> str:
    return format(float(value), ".6g")


def _format_python_value(value: Any) -> str:
    value = _coerce_snapshot_value(value)
    if isinstance(value, bool):
        return "True" if value else "False"
    if isinstance(value, (int, np.integer)) and not isinstance(value, bool):
        return str(int(value))
    if isinstance(value, (float, np.floating)):
        return _format_python_float(float(value))
    if isinstance(value, str):
        return repr(value)
    if isinstance(value, list):
        return "[" + ", ".join(_format_python_value(item) for item in value) + "]"
    if isinstance(value, tuple):
        inner = ", ".join(_format_python_value(item) for item in value)
        if len(value) == 1:
            inner += ","
        return f"({inner})"
    if value is None:
        return "None"
    return repr(value)


def _projection_label(perspective: bool) -> str:
    return PROJECTION_LABELS[0] if perspective else PROJECTION_LABELS[1]


def format_camera_summary(camera: Camera, *, topovec_name: str = "tv") -> str:
    """Render the current camera as executable Python code."""
    theta, phi, alpha = camera.get_orientation_angles()
    look_at = ", ".join(
        _format_python_value(v)
        for v in np.asarray(camera.look_at, dtype=float)
    )
    return (
        f"{topovec_name}.mgl.Camera.from_angles("
        f"theta={_format_python_value(theta)}, "
        f"phi={_format_python_value(phi)}, "
        f"alpha={_format_python_value(alpha)}, "
        f"scale={_format_python_value(camera.look_scale)}, "
        f"perspective={camera.perspective_is_on}, "
        f"look_at=[{look_at}])"
    )


def pil_image_to_data_url(image: Image.Image, image_format: str = "PNG") -> str:
    """Encode a PIL image for direct use in browser widget payloads."""
    buffer = io.BytesIO()
    image.save(buffer, format=image_format)
    payload = base64.b64encode(buffer.getvalue()).decode("ascii")
    mime = f"image/{image_format.lower()}"
    return f"data:{mime};base64,{payload}"


def apply_camera_command(
    camera: Camera,
    *,
    event_type: str,
    dx: float = 0.0,
    dy: float = 0.0,
    wheel: float = 0.0,
    rotate_rate: float = 2 * np.pi,
    roll_rate: float = 0.02 * np.pi,
    pan_rate: float = 2.0,
    zoom_rate: float = 0.12,
) -> Camera:
    """Apply one normalized interaction event to a camera.

    This helper is shared by widget event handlers and other adapters that want
    to reuse the same interaction semantics without depending on marimo.
    """
    dx = float(dx)
    dy = float(dy)
    wheel = float(wheel)

    if event_type == "rotate":
        camera.rotate_field_of_view(x=-dx, y=dy, rate=rotate_rate)
    elif event_type == "pan":
        camera.move_field_of_view(x=-dx, y=dy, rate=pan_rate)
    elif event_type == "roll":
        camera.rotate_around(z=-1.0, rate=dx * roll_rate)
    elif event_type == "zoom":
        camera.look_scale *= float(np.exp(np.clip(wheel, -8.0, 8.0) * zoom_rate))
    elif event_type in {"", "noop"}:
        return camera
    else:
        raise ValueError(f"Unsupported camera event: {event_type!r}")
    return camera


def _coerce_snapshot_value(value: Any) -> Any:
    if isinstance(value, np.generic):
        return value.item()
    if isinstance(value, np.ndarray):
        return value.tolist()
    if isinstance(value, tuple):
        return [_coerce_snapshot_value(v) for v in value]
    if isinstance(value, list):
        return [_coerce_snapshot_value(v) for v in value]
    return value


def _copy_shared_value(value: Any) -> Any:
    return copy.deepcopy(_coerce_snapshot_value(value))


def _shared_values_equal(left: Any, right: Any) -> bool:
    if isinstance(left, np.ndarray) or isinstance(right, np.ndarray):
        return np.array_equal(np.asarray(left), np.asarray(right))
    try:
        return bool(left == right)
    except Exception:
        return False


def _coerce_director_field_state(state: Any) -> np.ndarray:
    array = np.asarray(state)
    if array.ndim < 4:
        raise ValueError(
            "`inspect` expects a director field with at least three spatial "
            f"dimensions and a trailing vector axis, got shape {array.shape!r}."
        )
    if array.shape[-1] != 3:
        raise ValueError(
            "`inspect` expects the last state axis to have length 3, "
            f"got shape {array.shape!r}."
        )
    return array


def _infer_cubic_system_from_state(state: np.ndarray) -> System:
    return System.cubic(size=tuple(int(v) for v in state.shape[:3]))


def _attach_live_refs(target: Any, **refs: Any) -> Any:
    for name, value in refs.items():
        try:
            setattr(target, name, value)
        except Exception:
            continue
    return target


def _set_widget_frame_payload(target: Any, payload: dict[str, Any]) -> None:
    target.image_data = str(payload["image_data"])
    target.status_text = str(payload["status_text"])
    target.projection_mode = str(payload["projection_mode"])
    target.camera_state = dict(payload["camera_state"])
    target.width = int(payload.get("width", getattr(target, "width", 512)))
    target.height = int(payload.get("height", getattr(target, "height", 512)))


def _apply_bound_session_message(
    session: "RenderSession",
    message_type: str,
    content: dict[str, Any],
) -> _SessionMessageOutcome:
    if message_type == "camera_delta":
        changed = session.apply_camera_delta(
            event_type=str(content.get("event_type") or ""),
            dx=float(content.get("dx") or 0.0),
            dy=float(content.get("dy") or 0.0),
            wheel=float(content.get("wheel") or 0.0),
        )
        return _SessionMessageOutcome(
            handled=True,
            publish_frame=changed,
            publish_camera=changed,
            publish_summary=changed,
        )

    if message_type == "camera_commit":
        changed = session.commit_camera_interaction()
        return _SessionMessageOutcome(
            handled=True,
            publish_frame=changed,
            publish_camera=changed,
            publish_summary=changed,
        )

    if message_type == "toggle_perspective":
        changed = session.apply_camera_button("toggle_perspective")
        return _SessionMessageOutcome(
            handled=True,
            publish_frame=changed,
            publish_camera=changed,
            publish_summary=changed,
        )

    if message_type == "reset_camera":
        changed = session.apply_camera_button("reset")
        return _SessionMessageOutcome(
            handled=True,
            publish_frame=changed,
            publish_camera=changed,
            publish_summary=changed,
        )

    if message_type == "snap_camera":
        changed = session.apply_camera_button("snap")
        return _SessionMessageOutcome(
            handled=True,
            publish_frame=changed,
            publish_camera=changed,
        )

    if message_type == "switch_scene":
        residency_manager = getattr(session, "_scene_residency_manager", None)
        if residency_manager is None:
            changed = session.switch_scene(str(content.get("scene_name") or ""))
        else:
            changed = residency_manager.switch_scene(str(content.get("scene_name") or ""))
        return _SessionMessageOutcome(
            handled=True,
            publish_frame=changed,
            publish_camera=changed,
            publish_properties=changed,
            publish_summary=changed,
        )

    if message_type == "apply_property_change":
        session.apply_property_input(
            str(content.get("group_id") or ""),
            int(content.get("prop_index") or 0),
            content.get("value"),
            expected_field=(
                None
                if content.get("expected_field") is None
                else str(content.get("expected_field"))
            ),
            expected_title=(
                None
                if content.get("expected_title") is None
                else str(content.get("expected_title"))
            ),
        )
        return _SessionMessageOutcome(
            handled=True,
            publish_frame=True,
            publish_camera=True,
            publish_properties=True,
            publish_summary=True,
        )

    return _SessionMessageOutcome(handled=False)


def _property_snapshot(prop: Property, index: int) -> dict[str, Any]:
    spec = dict(prop.spec())
    snapshot = {
        "index": int(index),
        "kind": str(spec["kind"]),
        "field": str(spec["field"]),
        "title": str(spec["title"]),
        "value": _coerce_snapshot_value(spec["value"]),
        "invalidates": bool(spec["invalidates"]),
        "param_key": spec.get("param_key"),
    }
    if isinstance(prop, NumericProperty):
        snapshot.update(
            {
                "min": float(spec["min"]),
                "max": float(spec["max"]),
                "count": int(spec["count"]),
            }
        )
    elif isinstance(prop, ListProperty):
        snapshot.update(
            {
                "values": [_coerce_snapshot_value(v) for v in spec["values"]],
                "selected_index": int(spec["selected_index"]),
            }
        )
    return snapshot


def _numeric_step(prop_snapshot: dict[str, Any]) -> float | None:
    count = int(prop_snapshot.get("count", 0))
    minimum = float(prop_snapshot.get("min", 0.0))
    maximum = float(prop_snapshot.get("max", 0.0))
    if count <= 1:
        return None
    delta = maximum - minimum
    if not np.isfinite(delta) or delta <= 0:
        return None
    step = delta / float(count - 1)
    if not np.isfinite(step) or step <= 0:
        return None
    return float(step)


def _default_numeric_step() -> float:
    return DEFAULT_NUMERIC_STEP


def _effective_numeric_step(prop_snapshot: dict[str, Any]) -> float:
    step = _numeric_step(prop_snapshot)
    if step is not None:
        return step
    return _default_numeric_step()


def _numeric_bounds(
    prop_snapshot: dict[str, Any],
) -> tuple[float, float, float, bool]:
    """Normalize numeric metadata and report whether the current value fits it."""
    minimum = float(prop_snapshot.get("min", 0.0))
    maximum = float(prop_snapshot.get("max", 0.0))
    value = float(prop_snapshot.get("value", 0.0))

    if np.isfinite(minimum) and np.isfinite(maximum) and minimum > maximum:
        minimum, maximum = maximum, minimum

    value_in_bounds = (
        np.isfinite(minimum)
        and np.isfinite(maximum)
        and minimum <= value <= maximum
    )
    return minimum, maximum, value, value_in_bounds


def _numeric_step_decimals(step: float | None) -> int | None:
    if step is None:
        return None
    step = float(abs(step))
    if not np.isfinite(step) or step <= 0:
        return None
    text = f"{step:.12f}".rstrip("0").rstrip(".")
    if "." not in text:
        return 0
    return len(text.split(".", 1)[1])


def _quantize_numeric_value(value: float, prop_snapshot: dict[str, Any]) -> float:
    """Snap numeric input to either count-based or fallback resolution."""
    step = _effective_numeric_step(prop_snapshot)
    minimum, maximum, _, value_in_bounds = _numeric_bounds(prop_snapshot)
    snapped = float(value)

    if value_in_bounds:
        snapped = minimum + round((snapped - minimum) / step) * step
        snapped = float(np.clip(snapped, minimum, maximum))
    else:
        snapped = round(snapped / step) * step

    decimals = _numeric_step_decimals(step)
    if decimals is not None:
        snapped = round(snapped, decimals)

    return float(snapped)


def _decorate_property_snapshot(prop_snapshot: dict[str, Any]) -> dict[str, Any]:
    snapshot = dict(prop_snapshot)
    kind = str(snapshot.get("kind", ""))

    if kind == "numeric":
        minimum, maximum, value, value_in_bounds = _numeric_bounds(snapshot)
        step = _effective_numeric_step(snapshot)
        snapshot["value"] = _quantize_numeric_value(value, snapshot)
        snapshot["step"] = step
        snapshot["value_in_bounds"] = bool(value_in_bounds)
        snapshot["input_kind"] = (
            "slider"
            if step is not None and maximum > minimum and value_in_bounds
            else "number"
        )
    elif kind == "boolean":
        snapshot["input_kind"] = "switch"
    elif kind == "list":
        snapshot["input_kind"] = "dropdown"
    else:
        snapshot["input_kind"] = "display"

    snapshot["row_kind"] = (
        "stack" if snapshot["input_kind"] == "slider" else "inline"
    )

    return snapshot


def _missing_marimo_stack_error(
    missing: tuple[str, ...] = ("anywidget", "traitlets"),
) -> ImportError:
    missing_text = ", ".join(missing)
    return ImportError(
        "topovec.marimo.camera_viewer requires the optional widget stack "
        f"({missing_text}) in the active uv environment. "
        "Run `uv sync --all-extras` or `uv sync --extra marimo --extra mgl`, "
        "then restart or re-run the notebook cell."
    )


def _missing_marimo_error() -> ImportError:
    return ImportError(
        "topovec.marimo.camera_viewer requires `marimo` in the active uv environment. "
        "Run `uv sync --all-extras` or `uv sync --extra marimo --extra mgl`."
    )


def _load_marimo():
    try:
        import marimo as mo
    except ModuleNotFoundError as exc:
        raise _missing_marimo_error() from exc
    return mo


def _load_widget_stack():
    missing = []
    try:
        import anywidget
    except ModuleNotFoundError:
        anywidget = None
        missing.append("anywidget")
    try:
        import traitlets
    except ModuleNotFoundError:
        traitlets = None
        missing.append("traitlets")
    if missing:
        raise _missing_marimo_stack_error(tuple(missing))
    return anywidget, traitlets


@functools.lru_cache(maxsize=1)
def _camera_viewport_widget_class():
    """Create and cache the anywidget-backed viewport class."""
    anywidget, traitlets = _load_widget_stack()

    class _CameraViewportWidget(anywidget.AnyWidget):
        _esm = camera_viewport_esm()
        _css = camera_viewport_css()

        image_data = traitlets.Unicode("").tag(sync=True)
        status_text = traitlets.Unicode("").tag(sync=True)
        controls_text = traitlets.Unicode(DEFAULT_CONTROLS_TEXT).tag(sync=True)
        projection_mode = traitlets.Unicode(PROJECTION_LABELS[0]).tag(sync=True)
        camera_state = traitlets.Dict(default_value={}).tag(sync=True)
        width = traitlets.Int(512).tag(sync=True)
        height = traitlets.Int(512).tag(sync=True)
        busy = traitlets.Bool(False).tag(sync=True)
        show_toolbar = traitlets.Bool(True).tag(sync=True)
        use_custom_transport = traitlets.Bool(False).tag(sync=True)
        event_id = traitlets.Int(0).tag(sync=True)
        event_type = traitlets.Unicode("").tag(sync=True)
        event_dx = traitlets.Float(0.0).tag(sync=True)
        event_dy = traitlets.Float(0.0).tag(sync=True)
        event_wheel = traitlets.Float(0.0).tag(sync=True)

        def __init__(
            self,
            image_container: Any,
            state: np.ndarray | None = None,
            *,
            controls_text: str = DEFAULT_CONTROLS_TEXT,
            rotate_rate: float = 2 * np.pi,
            pan_rate: float = 2.0,
            zoom_rate: float = 0.12,
            show_toolbar: bool = True,
            auto_apply_events: bool = True,
            use_custom_transport: bool = False,
        ):
            """Bind an image container and initialize the first rendered frame."""
            super().__init__()
            self._image_container = image_container
            self._camera = image_container.scene.camera
            self._rotate_rate = float(rotate_rate)
            self._pan_rate = float(pan_rate)
            self._zoom_rate = float(zoom_rate)
            self._auto_apply_events = bool(auto_apply_events)
            self._bound_session: RenderSession | None = None
            self._publish_camera_snapshot = None
            self._publish_properties_snapshot = None
            self._last_payload: dict[str, Any] | None = None
            self.controls_text = controls_text
            self.show_toolbar = bool(show_toolbar)
            self.use_custom_transport = bool(use_custom_transport)

            size = getattr(
                image_container,
                "size",
                getattr(image_container, "_size", (512, 512)),
            )
            self.width = int(size[0])
            self.height = int(size[1])

            if state is not None:
                self._image_container.upload(state)

            self._home_state = serialize_camera(self._camera)
            self._register_message_handler()
            self.set_frame_payload(self._render_frame_payload())

        def _register_message_handler(self) -> None:
            on_msg = getattr(self, "on_msg", None)
            if callable(on_msg):
                on_msg(self._handle_custom_message)

        def bind_session(
            self,
            session: RenderSession,
            *,
            publish_camera_snapshot: Any | None = None,
            publish_properties_snapshot: Any | None = None,
        ) -> "_CameraViewportWidget":
            self._bound_session = session
            self.use_custom_transport = True
            if publish_camera_snapshot is not None:
                self._publish_camera_snapshot = publish_camera_snapshot
            if publish_properties_snapshot is not None:
                self._publish_properties_snapshot = publish_properties_snapshot
            return self

        def _publish_camera(self) -> None:
            if self._publish_camera_snapshot is not None and self._bound_session is not None:
                self._publish_camera_snapshot(self._bound_session.camera_snapshot())

        def _publish_properties(self) -> None:
            if (
                self._publish_properties_snapshot is not None
                and self._bound_session is not None
            ):
                self._publish_properties_snapshot(
                    self._bound_session.properties_snapshot()
                )

        def _ack_last_payload(self) -> "_CameraViewportWidget":
            if self._last_payload is not None:
                return self.set_frame_payload(self._last_payload)
            if self._bound_session is not None:
                return self.set_frame_payload(
                    self._bound_session.frame_payload(
                        image_encoder=pil_image_to_data_url
                    )
                )
            return self.set_frame_payload(self._render_frame_payload())

        def _set_error_status(self, message: str) -> None:
            self.busy = True
            try:
                self.status_text = str(message)
            finally:
                self.busy = False

        def _render_frame_payload(self) -> dict[str, Any]:
            image = self._image_container.save()
            return {
                "image_data": pil_image_to_data_url(image),
                "status_text": format_camera_summary(self._camera),
                "projection_mode": _projection_label(self._camera.perspective_is_on),
                "camera_state": serialize_camera(self._camera),
                "width": int(self.width),
                "height": int(self.height),
            }

        def set_frame_payload(self, payload: dict[str, Any]) -> "_CameraViewportWidget":
            """Update browser-visible traits from a rendered session payload."""
            self._last_payload = dict(payload)
            self.busy = True
            try:
                _set_widget_frame_payload(self, payload)
            finally:
                self.busy = False
            return self

        def refresh(self) -> "_CameraViewportWidget":
            """Render the current scene/camera state back into the widget image."""
            return self.set_frame_payload(self._render_frame_payload())

        def set_render_state(self, state: np.ndarray) -> "_CameraViewportWidget":
            """Replace the rendered director/state field without resetting the camera."""
            self._image_container.upload(state)
            return self.refresh()

        def set_home(self) -> "_CameraViewportWidget":
            """Store the current camera pose as the reset target."""
            if self._bound_session is not None:
                self._bound_session.set_home()
            self._home_state = serialize_camera(self._camera)
            return self.refresh()

        def reset_camera(self) -> "_CameraViewportWidget":
            """Restore the saved home camera pose and redraw the frame."""
            if self._bound_session is not None:
                changed = self._bound_session.apply_camera_button("reset")
                if changed:
                    self.set_frame_payload(
                        self._bound_session.frame_payload(
                            image_encoder=pil_image_to_data_url
                        )
                    )
                    self._publish_camera()
                    return self
                return self._ack_last_payload()
            restore_camera(self._camera, self._home_state)
            return self.refresh()

        def set_camera_state(
            self, camera_state: dict[str, Any]
        ) -> "_CameraViewportWidget":
            """Replace the camera with a serialized state and redraw the frame."""
            restore_camera(self._camera, camera_state)
            return self.refresh()

        def _handle_session_message(self, message_type: str, content: dict[str, Any]) -> bool:
            if self._bound_session is None:
                return False
            outcome = _apply_bound_session_message(
                self._bound_session,
                message_type,
                content,
            )
            if not outcome.handled:
                return False
            if outcome.publish_frame:
                self.set_frame_payload(
                    self._bound_session.frame_payload(
                        image_encoder=pil_image_to_data_url
                    )
                )
            else:
                self._ack_last_payload()
            if outcome.publish_camera:
                self._publish_camera()
            if outcome.publish_properties:
                self._publish_properties()
            return True

        def _handle_custom_message(self, _widget, content, _buffers) -> None:
            if not isinstance(content, dict):
                return
            message_type = str(content.get("type") or "")
            if not message_type:
                return

            try:
                handled = self._handle_session_message(message_type, content)
                if not handled:
                    raise ValueError(f"Unsupported viewer message: {message_type!r}")
            except Exception as exc:
                self._set_error_status(f"Render failed: {exc}")

        @traitlets.observe("event_id")
        def _handle_event(self, change):
            """Consume one aggregated frontend event in standalone mode."""
            if not self._auto_apply_events:
                return
            if change["new"] == change["old"]:
                return

            try:
                if self.event_type == "reset":
                    restore_camera(self._camera, self._home_state)
                elif self.event_type == "snap":
                    self._camera.snap()
                elif self.event_type == "toggle_perspective":
                    self._camera.toggle_perspective()
                else:
                    apply_camera_command(
                        self._camera,
                        event_type=self.event_type,
                        dx=self.event_dx,
                        dy=self.event_dy,
                        wheel=self.event_wheel,
                        rotate_rate=self._rotate_rate,
                        pan_rate=self._pan_rate,
                        zoom_rate=self._zoom_rate,
                    )
                self.refresh()
            except Exception as exc:
                self.status_text = f"Render failed: {exc}"
                self.busy = False

    _CameraViewportWidget.__name__ = "CameraViewportWidget"
    _CameraViewportWidget.__qualname__ = "CameraViewportWidget"
    return _CameraViewportWidget


@functools.lru_cache(maxsize=1)
def _composite_director_field_widget_class():
    """Create and cache the self-contained anywidget inspector class."""
    anywidget, traitlets = _load_widget_stack()

    class _CompositeDirectorFieldWidget(anywidget.AnyWidget):
        _esm = director_field_inspect_esm()
        _css = director_field_inspect_css()

        image_data = traitlets.Unicode("").tag(sync=True)
        status_text = traitlets.Unicode("").tag(sync=True)
        controls_text = traitlets.Unicode(DEFAULT_CONTROLS_TEXT).tag(sync=True)
        projection_mode = traitlets.Unicode(PROJECTION_LABELS[0]).tag(sync=True)
        camera_state = traitlets.Dict(default_value={}).tag(sync=True)
        camera_snapshot = traitlets.Dict(default_value={}).tag(sync=True)
        properties_snapshot = traitlets.Dict(default_value={}).tag(sync=True)
        width = traitlets.Int(512).tag(sync=True)
        height = traitlets.Int(512).tag(sync=True)
        summary_rows = traitlets.Int(8).tag(sync=True)
        summary_text = traitlets.Unicode("").tag(sync=True)
        busy = traitlets.Bool(False).tag(sync=True)
        show_toolbar = traitlets.Bool(False).tag(sync=True)
        show_scene_selector = traitlets.Bool(True).tag(sync=True)

        def __init__(
            self,
            session: "RenderSession",
            *,
            controls_text: str = DEFAULT_CONTROLS_TEXT,
            show_scene_selector: bool = True,
            show_toolbar: bool = False,
            summary_rows: int = 8,
        ):
            super().__init__()
            self._bound_session = session
            self._image_container = session.image_container
            self._last_payload: dict[str, Any] | None = None
            self.controls_text = controls_text
            self.show_toolbar = bool(show_toolbar)
            self.show_scene_selector = bool(show_scene_selector)
            self.summary_rows = max(1, int(summary_rows))

            size = getattr(
                self._image_container,
                "size",
                getattr(self._image_container, "_size", (512, 512)),
            )
            self.width = int(size[0])
            self.height = int(size[1])

            self._register_message_handler()
            self.refresh()

        def _register_message_handler(self) -> None:
            on_msg = getattr(self, "on_msg", None)
            if callable(on_msg):
                on_msg(self._handle_custom_message)

        def set_frame_payload(
            self,
            payload: dict[str, Any],
        ) -> "_CompositeDirectorFieldWidget":
            self._last_payload = dict(payload)
            self.busy = True
            try:
                _set_widget_frame_payload(self, payload)
            finally:
                self.busy = False
            return self

        def set_camera_snapshot(
            self,
            snapshot: dict[str, Any],
        ) -> "_CompositeDirectorFieldWidget":
            self.camera_snapshot = dict(snapshot)
            return self

        def set_properties_snapshot(
            self,
            snapshot: dict[str, Any],
        ) -> "_CompositeDirectorFieldWidget":
            self.properties_snapshot = dict(snapshot)
            return self

        def set_summary_text(self, summary_text: str) -> "_CompositeDirectorFieldWidget":
            self.summary_text = str(summary_text)
            return self

        def refresh(self) -> "_CompositeDirectorFieldWidget":
            return self._publish_state(
                publish_frame=True,
                publish_camera=True,
                publish_properties=True,
                publish_summary=True,
            )

        def _publish_state(
            self,
            *,
            publish_frame: bool = False,
            publish_camera: bool = False,
            publish_properties: bool = False,
            publish_summary: bool = False,
        ) -> "_CompositeDirectorFieldWidget":
            if publish_frame:
                self.set_frame_payload(
                    self._bound_session.frame_payload(
                        image_encoder=pil_image_to_data_url
                    )
                )
            if publish_camera:
                self.set_camera_snapshot(self._bound_session.camera_snapshot())
            if publish_properties:
                self.set_properties_snapshot(self._bound_session.properties_snapshot())
            if publish_summary:
                self.set_summary_text(self._bound_session.render_snippet())
            return self

        def _set_error_status(self, message: str) -> None:
            self.busy = True
            try:
                self.status_text = str(message)
            finally:
                self.busy = False

        def _handle_custom_message(self, _widget, content, _buffers) -> None:
            if not isinstance(content, dict):
                return
            message_type = str(content.get("type") or "")
            if not message_type:
                return

            try:
                outcome = _apply_bound_session_message(
                    self._bound_session,
                    message_type,
                    content,
                )
                if not outcome.handled:
                    raise ValueError(
                        f"Unsupported composite viewer message: {message_type!r}"
                    )
                self._publish_state(
                    publish_frame=outcome.publish_frame,
                    publish_camera=outcome.publish_camera,
                    publish_properties=outcome.publish_properties,
                    publish_summary=outcome.publish_summary,
                )
            except Exception as exc:
                self._set_error_status(f"Render failed: {exc}")

    _CompositeDirectorFieldWidget.__name__ = "CompositeDirectorFieldWidget"
    _CompositeDirectorFieldWidget.__qualname__ = "CompositeDirectorFieldWidget"
    return _CompositeDirectorFieldWidget


def _load_builtin_scene_registry() -> dict[str, Any]:
    try:
        from ..mgl import SCENES
    except Exception:
        return {}
    return dict(SCENES)


serialize_camera = _viewer_session_core.serialize_camera
restore_camera = _viewer_session_core.restore_camera
camera_state_signature = _viewer_session_core.camera_state_signature
format_camera_summary = _viewer_session_core.format_camera_summary
apply_camera_command = _viewer_session_core.apply_camera_command
_coerce_director_field_state = _viewer_session_core._coerce_director_field_state
_infer_cubic_system_from_state = _viewer_session_core.infer_cubic_system_from_state
_property_snapshot = _viewer_session_core._property_snapshot
_effective_numeric_step = _viewer_session_core._effective_numeric_step
_quantize_numeric_value = _viewer_session_core._quantize_numeric_value
_decorate_property_snapshot = _viewer_session_core._decorate_property_snapshot


class RenderSession(_viewer_session_core.ViewerSession):
    def _builtin_scene_registry(self) -> dict[str, Any]:
        return _load_builtin_scene_registry()


def camera_viewer(
    image_container: Any,
    state: np.ndarray | None = None,
    *,
    controls_text: str = DEFAULT_CONTROLS_TEXT,
    rotate_rate: float = 2 * np.pi,
    pan_rate: float = 2.0,
    zoom_rate: float = 0.12,
):
    """Create a standalone marimo viewport for an existing image container.

    This is the lowest-level marimo primitive in the module. It exposes camera
    interaction but does not build the full property inspector UI.
    """
    mo = _load_marimo()
    widget = _camera_viewport_widget_class()(
        image_container=image_container,
        state=state,
        controls_text=controls_text,
        rotate_rate=rotate_rate,
        pan_rate=pan_rate,
        zoom_rate=zoom_rate,
        show_toolbar=True,
        auto_apply_events=True,
    )
    return mo.ui.anywidget(widget)

def inspect_view(
    state: np.ndarray,
    *,
    system: System | None = None,
    scene: str = "Layer",
    imgsize: int = 512,
    scalex: bool = False,
    camera_preset: int | None = None,
    show_scene_selector: bool = True,
    show_toolbar: bool = False,
    summary_rows: int = 8,
    controls_text: str = DEFAULT_CONTROLS_TEXT,
    rotate_rate: float = 2 * np.pi,
    roll_rate: float = 0.02 * np.pi,
    pan_rate: float = 2.0,
    zoom_rate: float = 0.12,
) -> DirectorFieldView:
    """Build a self-contained marimo inspector for a director field.

    The helper prepares a default `topovec.mgl` scene, uploads the state, and
    returns a live bundle whose widget mirrors the underlying
    :class:`RenderSession`.
    """
    import topovec.mgl as mgl

    mo = _load_marimo()
    director_state = _coerce_director_field_state(state)
    resolved_system = (
        _infer_cubic_system_from_state(director_state)
        if system is None
        else system
    )
    image_container = mgl.render_prepare(
        scenecls=scene,
        system=resolved_system,
        camera_preset=camera_preset,
        scalex=scalex,
        imgsize=imgsize,
    )
    scene_residency = SceneResidencyManager(
        image_container,
        scene_registry_loader=_load_builtin_scene_registry,
        session_factory=RenderSession,
        session_kwargs={
            "controls_text": controls_text,
            "rotate_rate": rotate_rate,
            "roll_rate": roll_rate,
            "pan_rate": pan_rate,
            "zoom_rate": zoom_rate,
        },
        max_live_scenes=1,
    )
    session = scene_residency.activate(
        system=resolved_system,
        state=director_state,
        preferred_scene_name=scene,
    )
    raw_widget = _composite_director_field_widget_class()(
        session,
        controls_text=controls_text,
        show_scene_selector=show_scene_selector,
        show_toolbar=show_toolbar,
        summary_rows=summary_rows,
    )
    widget = mo.ui.anywidget(raw_widget)
    view = DirectorFieldView(
        session=session,
        widget=widget,
        layout=widget,
    )
    _attach_live_refs(
        widget,
        _topovec_director_field_view=view,
        _topovec_scene_residency=scene_residency,
    )
    return view


def inspect(
    state: np.ndarray,
    *,
    system: System | None = None,
    scene: str = "Layer",
    imgsize: int = 512,
    scalex: bool = False,
    camera_preset: int | None = None,
    show_scene_selector: bool = True,
    show_toolbar: bool = False,
    summary_rows: int = 8,
    controls_text: str = DEFAULT_CONTROLS_TEXT,
    rotate_rate: float = 2 * np.pi,
    roll_rate: float = 0.02 * np.pi,
    pan_rate: float = 2.0,
    zoom_rate: float = 0.12,
) -> Any:
    """Return only the final marimo widget for the default inspector view."""
    view = inspect_view(
        state,
        system=system,
        scene=scene,
        imgsize=imgsize,
        scalex=scalex,
        camera_preset=camera_preset,
        show_scene_selector=show_scene_selector,
        show_toolbar=show_toolbar,
        summary_rows=summary_rows,
        controls_text=controls_text,
        rotate_rate=rotate_rate,
        roll_rate=roll_rate,
        pan_rate=pan_rate,
        zoom_rate=zoom_rate,
    )
    return _attach_live_refs(
        view.layout,
        _topovec_director_field_view=view,
    )


def CameraViewportWidget(*args, **kwargs):
    """Compatibility factory returning the standalone viewport widget."""
    return _camera_viewport_widget_class()(*args, **kwargs)



import json
from types import SimpleNamespace

import numpy as np
import pytest
from PIL import Image

from topovec.core import BooleanProperty, ListProperty, NumericProperty, System, UpdateMask
from topovec.mgl.camera import Camera
from topovec.viewer.residency import SceneResidencyManager
from topovec.viewer.session import (
    ViewerSession,
    ViewerSessionResidencySnapshot,
    _decorate_property_snapshot,
    _effective_numeric_step,
    camera_state_signature,
)


class DummySystem:
    def __init__(self, dim=3):
        self.dim = dim


class DummySceneBase:
    def __init__(self, mglctx=None, system=None, camera=None):
        self._mglctx = mglctx
        self._system = DummySystem() if system is None else system
        self.camera = Camera() if camera is None else camera
        self._cameras = {str(self.camera_slot_key): self.camera}
        self.uploaded_state = None
        self.list_properties_calls = 0
        self.release_count = 0
        self.activate_camera_slot()

    @property
    def system(self):
        return self._system

    @property
    def camera_slot_key(self):
        return self.camera_profile

    @property
    def camera_profile(self):
        return "general"

    def _make_default_camera(self, slot):
        return Camera()

    def activate_camera_slot(self):
        slot = str(self.camera_slot_key)
        if slot not in self._cameras:
            self._cameras[slot] = self._make_default_camera(slot)
        self.camera = self._cameras[slot]
        return self.camera

    @classmethod
    def supports_system(cls, system):
        return True

    def upload(self, state=None):
        self.uploaded_state = state

    def release(self):
        self.release_count += 1


class DummyScene(DummySceneBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.flag = False
        self.level = 0.5
        self._axis = 2
        self.axis = 2
        self._mode = "alpha"
        self._min_level = 0.0
        self._max_level = 1.0

    @property
    def camera_profile(self):
        return f"layer-{('x', 'y', 'z')[getattr(self, '_axis', 2)]}"

    def _make_default_camera(self, slot):
        camera = Camera()
        if slot == "layer-x":
            camera.set_basis(depth=[1, 0, 0], up=[0, 0, 1])
        elif slot == "layer-y":
            camera.set_basis(depth=[0, 1, 0], up=[1, 0, 0])
        else:
            camera.set_basis(depth=[0, 0, 1], up=[1, 0, 0])
        return camera

    def get_mode(self):
        return self._mode

    def set_mode(self, value):
        self._mode = value
        if value == "alpha":
            self._min_level = 0.0
            self._max_level = 1.0
        else:
            self._min_level = -2.0
            self._max_level = 2.0
        self.level = float(np.clip(self.level, self._min_level, self._max_level))
        return UpdateMask.REDRAW | UpdateMask.REBUILD_PROPERTIES

    mode = property(get_mode, set_mode)

    def get_axis(self):
        return self._axis

    def set_axis(self, value):
        self._axis = int(value)
        return (
            UpdateMask.REDRAW
            | UpdateMask.REBUILD_PROPERTIES
            | UpdateMask.CAMERA_SLOT_CHANGED
        )

    axis = property(get_axis, set_axis)

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "Render": [
                ListProperty(
                    self,
                    "mode",
                    title="Mode",
                    values=["alpha", "beta"],
                    reset=True,
                ),
                NumericProperty(
                    self,
                    "level",
                    title="Level",
                    min=self._min_level,
                    max=self._max_level,
                    count=5,
                    param_key="shared.level",
                ),
                BooleanProperty(
                    self,
                    "flag",
                    title="Flag",
                    param_key="shared.flag",
                ),
                NumericProperty(
                    self,
                    "axis",
                    title="Axis",
                    min=0,
                    max=2,
                    count=3,
                    reset=True,
                    param_key="layer.axis",
                ),
            ]
        }


class DummyBlochScene(DummySceneBase):
    @property
    def camera_profile(self):
        return "bloch"

    def _make_default_camera(self, slot):
        camera = Camera()
        camera.set_basis(depth=[0, 1, 1], up=[1, 0, 0])
        return camera

    def list_properties(self):
        self.list_properties_calls += 1
        return {"Ball": []}


class DummyVectorScene(DummySceneBase):
    SHAPES = ["Cone", "Double cone", "Cylinder", "Arrow"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mesh_shape = "Cone"

    @property
    def camera_profile(self):
        return "vector"

    def get_mesh_shape(self):
        return self._mesh_shape

    def set_mesh_shape(self, value):
        if value not in self.SHAPES:
            raise ValueError(f"Unsupported mesh shape: {value!r}")
        self._mesh_shape = str(value)
        return UpdateMask.REDRAW

    mesh_shape = property(get_mesh_shape, set_mesh_shape)

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "Vector": [
                ListProperty(
                    self,
                    "mesh_shape",
                    title="Shape",
                    values=list(self.SHAPES),
                    reset=True,
                )
            ]
        }


class DummyScopedVectorScene(DummySceneBase):
    SHAPES = ["Cone", "Double cone", "Cylinder", "Arrow"]

    def __init__(self, *args, mutation_state=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._mesh_shape = "Cone"
        self._mutation_state = mutation_state if mutation_state is not None else {}

    def get_mesh_shape(self):
        return self._mesh_shape

    def set_mesh_shape(self, value):
        if not self._mutation_state.get("active", False):
            raise RuntimeError("scene mutation scope is not active")
        if value not in self.SHAPES:
            raise ValueError(f"Unsupported mesh shape: {value!r}")
        self._mesh_shape = str(value)
        return UpdateMask.REDRAW

    mesh_shape = property(get_mesh_shape, set_mesh_shape)

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "Vector": [
                ListProperty(
                    self,
                    "mesh_shape",
                    title="Shape",
                    values=list(self.SHAPES),
                    reset=True,
                    param_key="layer.mesh_shape",
                )
            ]
        }


class DummySchemaScene(DummySceneBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mode = "alpha"

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "Render": [
                ListProperty(
                    self,
                    "mode",
                    title="Mode",
                    values=["alpha", "beta"],
                    reset=True,
                )
            ]
        }


class DummyStyleScene(DummySceneBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.white_background = True

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "Style": [
                BooleanProperty(
                    self,
                    "white_background",
                    title="White background",
                    param_key="style.white_background",
                )
            ]
        }


class DummyStyleVariantScene(DummySceneBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.white_background = True
        self.flag = False

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "Style": [
                BooleanProperty(
                    self,
                    "white_background",
                    title="White background",
                    param_key="style.white_background",
                ),
                BooleanProperty(
                    self,
                    "flag",
                    title="Flag",
                ),
            ]
        }


class DummyImageContainer:
    def __init__(self, size=(32, 24), scene_factory=None):
        self._size = size
        self._scene_factory = scene_factory
        self.scene = None
        self.uploaded_state = None
        self.saved_frames = 0
        self.attach_scene(scene_factory or (lambda _ctx: DummyScene()))

    @property
    def size(self):
        return self._size

    def attach_scene(self, scene_factory):
        self._scene_factory = scene_factory
        if callable(scene_factory):
            self.scene = scene_factory(None)
        else:
            self.scene = scene_factory
        return self

    def upload(self, state):
        self.uploaded_state = state
        if hasattr(self.scene, "upload"):
            self.scene.upload(state)
        return self

    def save(self):
        self.saved_frames += 1
        return Image.new(
            "RGB",
            self._size,
            color=((self.saved_frames * 23) % 255, 34, 56),
        )


class ScopedDummyHost:
    def __init__(self, scene):
        self._scene = scene
        self.scope_calls = []

    @property
    def scene(self):
        return self._scene

    @property
    def size(self):
        return (32, 24)

    def create_scene(self, scene_cls, system):
        return scene_cls(system=system)

    def attach_scene(self, scene):
        self._scene = scene
        return self

    def upload(self, state):
        if hasattr(self._scene, "upload"):
            self._scene.upload(state)
        return self

    def render_image(self, *, size=None):
        return Image.new("RGB", self.size, color=(12, 34, 56))

    def scene_mutation_scope(self, reason: str):
        self.scope_calls.append(f"enter:{reason}")
        self._scene._mutation_state["active"] = True

        class _Scope:
            def __enter__(inner_self):
                return self

            def __exit__(inner_self, exc_type, exc, tb):
                self._scene._mutation_state["active"] = False
                self.scope_calls.append(f"exit:{reason}")
                return False

        return _Scope()


def install_dummy_scene_registry(monkeypatch):
    registry = {
        "Layer": DummyScene,
        "Ball": DummyBlochScene,
    }
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: dict(registry),
    )
    return registry


def test_viewer_session_snapshot_contains_property_groups(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    snapshot = session.snapshot()

    assert snapshot["revision"] == 0
    assert snapshot["selected_scene"] == "Layer"
    assert snapshot["available_scenes"] == ["Layer", "Ball"]
    assert snapshot["property_groups"][0]["id"] == "Render"
    assert "ic = tv.mgl.render_prepare('Layer', system)" in snapshot["summary_text"]


def test_render_snippet_does_not_create_extra_baseline_scene(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    create_calls = []
    original_create_scene = ViewerSession._create_scene

    def counted_create_scene(self, scene_cls):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(self, scene_cls)

    monkeypatch.setattr(ViewerSession, "_create_scene", counted_create_scene)
    session = ViewerSession(DummyImageContainer())

    assert create_calls == []

    snippet = session.render_snippet()

    assert "render_prepare('Layer', system)" in snippet
    assert create_calls == []


def test_viewer_session_frame_payload_accepts_encoder(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    payload = session.frame_payload(image_encoder=lambda image: f"img:{image.size}")

    assert payload["image_data"] == "img:(32, 24)"
    assert payload["projection_mode"] == "Perspective"
    assert payload["width"] == 32
    assert payload["height"] == 24


def test_viewer_session_switch_scene_restores_camera_slots(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    layer_camera = session.snapshot()["camera"]
    session.apply_property_change(
        "Render",
        3,
        0.0,
        expected_field="axis",
        expected_title="Axis",
    )
    layer_x_camera = session.snapshot()["camera"]

    assert camera_state_signature(layer_camera) != camera_state_signature(layer_x_camera)

    assert session.switch_scene("Ball") is True
    ball_camera = session.snapshot()["camera"]
    assert session.switch_scene("Layer") is True
    restored_x = session.snapshot()["camera"]

    assert camera_state_signature(ball_camera) != camera_state_signature(restored_x)


def test_switch_scene_creates_only_live_scene_and_reuses_cache(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    create_calls = []
    original_create_scene = ViewerSession._create_scene

    def counted_create_scene(self, scene_cls):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(self, scene_cls)

    monkeypatch.setattr(ViewerSession, "_create_scene", counted_create_scene)
    session = ViewerSession(DummyImageContainer())

    assert session.switch_scene("Ball") is True
    assert create_calls == ["DummyBlochScene"]

    session.render_snippet()
    assert create_calls == ["DummyBlochScene"]

    assert session.switch_scene("Layer") is True
    assert create_calls == ["DummyBlochScene"]

    assert session.switch_scene("Ball") is True
    assert create_calls == ["DummyBlochScene"]


def test_viewer_session_residency_snapshot_roundtrip(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    assert session.switch_scene("Ball") is True
    snapshot = session.export_residency_snapshot()

    assert isinstance(snapshot, ViewerSessionResidencySnapshot)
    assert snapshot.selected_scene == "Ball"
    assert "Layer" in snapshot.default_property_values_by_scene
    assert "camera:bloch" in snapshot.shared_state

    restored = ViewerSession(
        DummyImageContainer(scene_factory=lambda _ctx: DummyBlochScene()),
    )
    restored.restore_residency_snapshot(snapshot)

    assert restored.export_residency_snapshot() == snapshot


def test_scene_residency_manager_reuses_live_selected_scene_and_preserves_state(
    monkeypatch,
):
    install_dummy_scene_registry(monkeypatch)
    host = DummyImageContainer()
    manager = SceneResidencyManager(
        host,
        scene_registry_loader=lambda: {
            "Layer": DummyScene,
            "Ball": DummyBlochScene,
        },
        session_factory=ViewerSession,
        max_live_scenes=2,
    )
    state_a0 = np.zeros((4, 4, 4, 1, 3), dtype=np.float32)
    state_a1 = np.ones((4, 4, 4, 1, 3), dtype=np.float32)
    state_b0 = np.full((5, 4, 4, 1, 3), 2.0, dtype=np.float32)

    session_a = manager.activate(
        system=System.cubic((4, 4, 4)),
        state=state_a0,
        preferred_scene_name="Layer",
    )
    assert session_a.switch_scene("Ball") is True
    session_a.camera.look_scale *= 1.75
    session_a.commit_camera_interaction()
    ball_scene = session_a.scene
    camera_a = camera_state_signature(session_a.camera_snapshot())

    session_b = manager.activate(
        system=System.cubic((5, 4, 4)),
        state=state_b0,
        preferred_scene_name="Layer",
    )
    assert session_b is not session_a

    session_a_reloaded = manager.activate(
        system=System.cubic((4, 4, 4)),
        state=state_a1,
        preferred_scene_name="Layer",
    )

    assert session_a_reloaded is not session_a
    assert session_a_reloaded.selected_scene == "Ball"
    assert session_a_reloaded.scene is ball_scene
    assert camera_state_signature(session_a_reloaded.camera_snapshot()) == camera_a


def test_scene_residency_manager_evicts_lru_non_active_scene(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    host = DummyImageContainer()
    manager = SceneResidencyManager(
        host,
        scene_registry_loader=lambda: {
            "Layer": DummyScene,
            "Ball": DummyBlochScene,
        },
        session_factory=ViewerSession,
        max_live_scenes=2,
    )

    session_a = manager.activate(
        system=System.cubic((4, 4, 4)),
        state=np.zeros((4, 4, 4, 1, 3), dtype=np.float32),
        preferred_scene_name="Layer",
    )
    layer_scene = session_a.scene
    assert manager.switch_scene("Ball") is True

    manager.activate(
        system=System.cubic((5, 4, 4)),
        state=np.ones((5, 4, 4, 1, 3), dtype=np.float32),
        preferred_scene_name="Layer",
    )

    assert layer_scene.release_count == 1


def test_scene_residency_manager_applies_vector_shape_property_change(monkeypatch):
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: {"Layer": DummyVectorScene},
    )
    host = DummyImageContainer()
    manager = SceneResidencyManager(
        host,
        scene_registry_loader=lambda: {"Layer": DummyVectorScene},
        session_factory=ViewerSession,
        max_live_scenes=1,
    )

    session = manager.activate(
        system=System.cubic((4, 4, 4)),
        state=np.zeros((4, 4, 4, 1, 3), dtype=np.float32),
        preferred_scene_name="Layer",
    )

    session.apply_property_change(
        "Vector",
        0,
        "Arrow",
        expected_field="mesh_shape",
        expected_title="Shape",
    )

    assert manager.session is session
    assert session.scene.mesh_shape == "Arrow"
    assert (
        session.properties_snapshot()["property_groups"][0]["properties"][0]["value"]
        == "Arrow"
    )
    payload = session.frame_payload(image_encoder=lambda image: image.size)
    assert payload["image_data"] == (32, 24)


def test_apply_property_change_returns_redraw_mask_for_simple_property(monkeypatch):
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: {"Layer": DummyVectorScene},
    )
    session = ViewerSession(DummyImageContainer(scene_factory=lambda _ctx: DummyVectorScene()))

    mask = session.apply_property_change(
        "Vector",
        0,
        "Arrow",
        expected_field="mesh_shape",
        expected_title="Shape",
    )

    assert mask == UpdateMask.REDRAW


def test_properties_snapshot_includes_default_metadata(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    props = session.properties_snapshot()["property_groups"][0]["properties"]
    mode_prop = props[0]
    level_prop = props[1]

    assert mode_prop["default_value"] == "alpha"
    assert mode_prop["is_default"] is True
    assert mode_prop["can_reset"] is True
    assert level_prop["default_value"] == 0.5
    assert level_prop["is_default"] is True
    assert level_prop["can_reset"] is True


def test_effective_numeric_step_keeps_raw_resolution():
    prop_snapshot = {
        "min": 0.0,
        "max": 1.0,
        "count": 360,
        "value": 0.0,
    }

    assert _effective_numeric_step(prop_snapshot) == pytest.approx(1.0 / 359.0)


@pytest.mark.parametrize(
    ("minimum", "maximum", "count", "expected_step"),
    [
        (0.0, 2.0, 1000, 0.002),
        (0.0, 1.0, 360, 0.0028),
        (0.01, 2.0, 100, 0.02),
        (0.0, 2.0, 30, 0.069),
        (0.0, 0.02, 1000, 0.00002),
    ],
)
def test_decorate_property_snapshot_normalizes_numeric_ui_step(
    minimum,
    maximum,
    count,
    expected_step,
):
    snapshot = _decorate_property_snapshot(
        {
            "kind": "numeric",
            "min": minimum,
            "max": maximum,
            "count": count,
            "value": minimum,
        }
    )

    assert snapshot["step"] == pytest.approx(expected_step)


def test_apply_property_input_clamps_numeric_without_quantizing(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    mask = session.apply_property_input(
        "Render",
        1,
        0.26,
        expected_field="level",
        expected_title="Level",
    )

    assert mask == UpdateMask.REDRAW
    assert session.scene.level == 0.26
    assert (
        session.properties_snapshot()["property_groups"][0]["properties"][1]["value"]
        == 0.26
    )


def test_apply_property_change_restores_default_value(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    session.apply_property_change(
        "Render",
        2,
        True,
        expected_field="flag",
        expected_title="Flag",
    )
    snapshot = session.properties_snapshot()["property_groups"][0]["properties"][2]

    assert snapshot["is_default"] is False

    session.apply_property_change(
        "Render",
        2,
        snapshot["default_value"],
        expected_field="flag",
        expected_title="Flag",
    )

    restored = session.properties_snapshot()["property_groups"][0]["properties"][2]
    assert session.scene.flag is False
    assert restored["is_default"] is True


def test_apply_property_change_returns_rebuild_mask_without_schema(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    mask = session.apply_property_change(
        "Render",
        0,
        "beta",
        expected_field="mode",
        expected_title="Mode",
    )

    assert mask & UpdateMask.REBUILD_PROPERTIES
    assert not (mask & UpdateMask.REBUILD_PROPERTY_SCHEMA)


def test_apply_property_change_returns_schema_mask_for_reset_property(monkeypatch):
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: {"Layer": DummySchemaScene},
    )
    session = ViewerSession(
        DummyImageContainer(scene_factory=lambda _ctx: DummySchemaScene())
    )

    mask = session.apply_property_change(
        "Render",
        0,
        "beta",
        expected_field="mode",
        expected_title="Mode",
    )

    assert mask & UpdateMask.REBUILD_PROPERTIES
    assert mask & UpdateMask.REBUILD_PROPERTY_SCHEMA


def test_white_background_snapshot_and_shared_state_persist_across_scene_switch(
    monkeypatch,
):
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: {
            "Layer": DummyStyleScene,
            "Variant": DummyStyleVariantScene,
        },
    )
    session = ViewerSession(DummyImageContainer(scene_factory=lambda _ctx: DummyStyleScene()))

    mask = session.apply_property_change(
        "Style",
        0,
        False,
        expected_field="white_background",
        expected_title="White background",
    )
    snapshot = session.properties_snapshot()

    assert mask == UpdateMask.REDRAW
    assert snapshot["property_groups"][0]["id"] == "Style"
    assert snapshot["property_groups"][0]["properties"][0]["param_key"] == "style.white_background"
    assert snapshot["property_groups"][0]["properties"][0]["value"] is False

    assert session.switch_scene("Variant") is True
    variant_props = session.properties_snapshot()["property_groups"][0]["properties"]
    fields = {prop["field"] for prop in variant_props}

    assert "white_background" in fields
    assert "inverted" not in fields
    assert variant_props[0]["value"] is False


def test_apply_property_change_preserves_camera_slot_changed_mask(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = ViewerSession(DummyImageContainer())

    mask = session.apply_property_change(
        "Render",
        3,
        0.0,
        expected_field="axis",
        expected_title="Axis",
    )

    assert mask & UpdateMask.CAMERA_SLOT_CHANGED


def test_viewer_session_applies_context_sensitive_property_under_mutation_scope():
    mutation_state = {"active": False}
    scene = DummyScopedVectorScene(mutation_state=mutation_state)
    host = ScopedDummyHost(scene)
    session = ViewerSession(host)

    session.apply_property_change(
        "Vector",
        0,
        "Arrow",
        expected_field="mesh_shape",
        expected_title="Shape",
    )

    assert session.scene.mesh_shape == "Arrow"
    assert host.scope_calls == [
        "enter:sync_active_scene_from_shared_state",
        "exit:sync_active_scene_from_shared_state",
        "enter:apply_property_change",
        "exit:apply_property_change",
    ]


def test_viewer_session_syncs_context_sensitive_shared_property_under_mutation_scope():
    mutation_state = {"active": False}
    scene = DummyScopedVectorScene(mutation_state=mutation_state)
    host = ScopedDummyHost(scene)
    session = ViewerSession(host)
    session._shared_state["layer.mesh_shape"] = "Cylinder"

    session._sync_active_scene_from_shared_state()

    assert session.scene.mesh_shape == "Cylinder"
    assert host.scope_calls[-2:] == [
        "enter:sync_active_scene_from_shared_state",
        "exit:sync_active_scene_from_shared_state",
    ]



from __future__ import annotations

import functools
from importlib import resources


_ASSET_PACKAGE = "topovec.marimo.assets"


@functools.lru_cache(maxsize=None)
def _read_asset_text(name: str) -> str:
    try:
        return resources.files(_ASSET_PACKAGE).joinpath(name).read_text(
            encoding="utf-8"
        )
    except (FileNotFoundError, ModuleNotFoundError) as exc:
        raise RuntimeError(
            f"Missing marimo widget asset {name!r} in package {_ASSET_PACKAGE!r}. "
            "The installed distribution may have been built without widget assets."
        ) from exc


def camera_viewport_esm() -> str:
    return _read_asset_text("camera_viewport.js")


def camera_viewport_css() -> str:
    return _read_asset_text("camera_viewport.css")


def director_field_inspect_esm() -> str:
    return _read_asset_text("director_field_inspect.js")


def director_field_inspect_css() -> str:
    return _read_asset_text("director_field_inspect.css")



"""GPU buffer adapters used by ModernGL renders.

These classes only translate numpy array layouts to GPU buffers and back. They
do not own scene semantics, camera state, or frontend behavior.
"""

import numpy as np
import moderngl as mgl

from ..core import System


class Field3DBuffer:
    """Upload a vector field with one 3-float record per lattice site."""

    def __init__(self, mglctx: mgl.Context = None, system: System = None):
        self._ctx = mglctx
        self._system = system
        self.vbo = self._ctx.buffer(reserve=12 * self._system.number_of_spins)

    def upload(self, state: np.ndarray):
        """Write a ``(..., 3)`` state array into the underlying VBO."""
        assert isinstance(state, np.ndarray)
        self.vbo.write(state.reshape((-1, 3)).astype("f4").tobytes())

    def release(self) -> None:
        """Release the underlying ModernGL buffer."""
        vbo = getattr(self, "vbo", None)
        if vbo is None:
            return
        try:
            vbo.release()
        except Exception:
            pass
        self.vbo = None


class Field1DBuffer:
    """Upload one float per lattice site."""

    def __init__(self, mglctx: mgl.Context = None, system: System = None):
        self._ctx = mglctx
        self._system = system
        self.vbo = self._ctx.buffer(reserve=4 * self._system.number_of_spins)

    def upload(self, state: np.ndarray):
        """Write a scalar field into the underlying VBO."""
        assert isinstance(state, np.ndarray)
        self.vbo.write(state.flatten().astype("f4").tobytes())

    def release(self) -> None:
        """Release the underlying ModernGL buffer."""
        vbo = getattr(self, "vbo", None)
        if vbo is None:
            return
        try:
            vbo.release()
        except Exception:
            pass
        self.vbo = None


class Float3DBuffer:
    """Fixed-size buffer for arbitrary 3-float records."""

    def __init__(self, mglctx: mgl.Context = None, count: int = None):
        self._ctx = mglctx
        self._count = count
        self.vbo = self._ctx.buffer(reserve=12 * self._count)

    def upload(self, data: np.ndarray):
        """Write exactly ``count`` 3-float records."""
        assert isinstance(data, np.ndarray)
        data = data.reshape((-1, 3))
        assert data.shape[0] == self._count
        self.vbo.write(data.astype("f4").tobytes())

    def download(self):
        """Read the buffer back as a ``(count, 3)`` float array."""
        return np.frombuffer(self.vbo.read(), dtype=np.float32).reshape(
            (-1, 3)
        )  # , count=self._count)

    def release(self) -> None:
        """Release the underlying ModernGL buffer."""
        vbo = getattr(self, "vbo", None)
        if vbo is None:
            return
        try:
            vbo.release()
        except Exception:
            pass
        self.vbo = None


class IntBuffer:
    """Fixed-size integer index or mask buffer."""

    def __init__(self, mglctx: mgl.Context = None, count=None):
        self._ctx = mglctx
        self._count = count
        self.ibo = self._ctx.buffer(reserve=4 * self._count)

    @property
    def count(self):
        return self._count

    def upload(self, data: np.ndarray):
        """Write exactly ``count`` int32 values."""
        assert isinstance(data, np.ndarray)
        data = data.flatten()
        assert data.shape == (self._count,)
        self.ibo.write(data.astype("i4").tobytes())

    def download(self):
        """Read the buffer back as an ``int32`` array."""
        return np.frombuffer(self.ibo.read(), dtype=np.int32)  # , count=self._count)

    def release(self) -> None:
        """Release the underlying ModernGL buffer."""
        ibo = getattr(self, "ibo", None)
        if ibo is None:
            return
        try:
            ibo.release()
        except Exception:
            pass
        self.ibo = None



"""Small vector-field helpers shared by ansatz implementations.

These helpers intentionally stay lightweight. They encode common array-shape
operations used throughout the ansatz package without carrying any geometric
state of their own.
"""

import numpy as np


def pack_vector(x, y, z) -> np.ndarray:
    """Stack three scalar arrays into a trailing-dimension-3 vector field."""

    return np.stack([x, y, z], axis=-1)


def normalize_vector(vector: np.ndarray, eps: float = 1e-32) -> np.ndarray:
    """Normalize a vector field pointwise along its trailing vector axis."""

    vector = np.asarray(vector)
    norm = np.sqrt(np.sum(vector**2, axis=-1, keepdims=True) + eps)
    return vector / norm



"""ModernGL-based rendering primitives and convenience entrypoints.

The package exposes reusable scene and camera abstractions, a headless image
container, and a small set of convenience factories for standard rendering
setups. Frontends such as `topovec.marimo` build on this package rather than
redefining rendering behavior.
"""

import importlib

__all__ = [
    "ImageContainer",
    "Camera",
    "LayerScene",
    "CubeScene",
    "FlowScene",
    "AstraScene",
    "LevelScene",
    "ChargeScene",
    "VectorScene",
    "IsolinesScene",
    "SphereDotScene",
    "LayerAndSurfaceScene",
    "LayerAndIsolinesScene",
    "SCENES",
    "print_scenes",
    "render_layer",
    "render_isosurface",
    "render_isolines",
    "render_prepare",
]

_EXPORTS = {
    "ImageContainer": (".image", "ImageContainer"),
    "Camera": (".camera", "Camera"),
    "LayerScene": (".scene", "LayerScene"),
    "CubeScene": (".scene", "CubeScene"),
    "FlowScene": (".scene", "FlowScene"),
    "AstraScene": (".scene", "AstraScene"),
    "LevelScene": (".scene", "LevelScene"),
    "ChargeScene": (".scene", "ChargeScene"),
    "VectorScene": (".scene", "VectorScene"),
    "IsolinesScene": (".scene", "IsolinesScene"),
    "SphereDotScene": (".scene", "SphereDotScene"),
    "LayerAndSurfaceScene": (".scene", "LayerAndSurfaceScene"),
    "LayerAndIsolinesScene": (".scene", "LayerAndIsolinesScene"),
    "SCENES": (".scene", "SCENES"),
    "print_scenes": (".scene", "print_scenes"),
    "render_layer": (".sugar", "render_layer"),
    "render_isosurface": (".sugar", "render_isosurface"),
    "render_isolines": (".sugar", "render_isolines"),
    "render_prepare": (".sugar", "render_prepare"),
}


def __getattr__(name):
    if name not in _EXPORTS:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    module_name, attr_name = _EXPORTS[name]
    module = importlib.import_module(module_name, __name__)
    value = getattr(module, attr_name)
    globals()[name] = value
    return value


def __dir__():
    return sorted(set(globals()) | set(__all__))



"""CLI helpers for the ``topovec view`` desktop viewer command."""

from __future__ import annotations

import argparse
import faulthandler
import logging
import os
import platform
import sys

from ...core.log import configLog

logger = logging.getLogger(__name__)

_INTERESTING_ENV_VARS = (
    "QT_QPA_PLATFORM",
    "QT_QPA_PLATFORMTHEME",
    "XDG_SESSION_TYPE",
    "XDG_CURRENT_DESKTOP",
    "DESKTOP_SESSION",
    "DISPLAY",
    "WAYLAND_DISPLAY",
    "EGL_PLATFORM",
    "DRI_PRIME",
    "__NV_PRIME_RENDER_OFFLOAD",
    "__GLX_VENDOR_LIBRARY_NAME",
    "__VK_LAYER_NV_optimus",
    "MESA_LOADER_DRIVER_OVERRIDE",
    "LIBGL_ALWAYS_SOFTWARE",
)


def _load_qt():
    try:
        from PySide6 import QtCore, QtGui, QtWidgets
    except ModuleNotFoundError as exc:
        raise ImportError(
            "topovec view requires the optional GUI stack. Install TopoVec with "
            "`uv tool install 'topovec[view]'`, `uv sync --extra view`, "
            "or `uv sync --all-extras`."
        ) from exc
    return QtCore, QtGui, QtWidgets


def _positive_int(value: str) -> int:
    parsed = int(value)
    if parsed < 1:
        raise argparse.ArgumentTypeError("--max-live-scenes must be at least 1.")
    return parsed


def add_view_arguments(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable verbose viewer diagnostics in the console and log dock.",
    )
    parser.add_argument(
        "--max-live-scenes",
        type=_positive_int,
        default=2,
        help="Maximum number of live scene objects to keep resident across viewer systems.",
    )
    parser.add_argument("paths", nargs="*", help="Files to open on startup.")
    return parser


def build_parser() -> argparse.ArgumentParser:
    return add_view_arguments(
        argparse.ArgumentParser(
            prog="topovec view",
            description="Desktop viewer for TopoVec NPZ and SAD state files.",
        )
    )


def _install_console_excepthook() -> object:
    previous = sys.excepthook

    def _hook(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            previous(exc_type, exc_value, exc_traceback)
            return
        logger.critical(
            "Unhandled exception in topovec view",
            exc_info=(exc_type, exc_value, exc_traceback),
        )
        previous(exc_type, exc_value, exc_traceback)

    sys.excepthook = _hook
    return previous


def _install_qt_message_logging(QtCore) -> object:
    level_by_type = {
        QtCore.QtMsgType.QtDebugMsg: logging.DEBUG,
        QtCore.QtMsgType.QtInfoMsg: logging.INFO,
        QtCore.QtMsgType.QtWarningMsg: logging.WARNING,
        QtCore.QtMsgType.QtCriticalMsg: logging.ERROR,
        QtCore.QtMsgType.QtFatalMsg: logging.CRITICAL,
    }

    def _handler(message_type, context, message):
        logger.log(level_by_type.get(message_type, logging.INFO), "Qt: %s", message)

    return QtCore.qInstallMessageHandler(_handler)


def _enum_name(enum_value) -> str:
    try:
        return str(enum_value).rsplit(".", 1)[-1]
    except Exception:
        return repr(enum_value)


def _surface_format_snapshot(fmt) -> dict[str, object]:
    return {
        "version": (int(fmt.majorVersion()), int(fmt.minorVersion())),
        "profile": _enum_name(fmt.profile()),
        "renderable_type": _enum_name(fmt.renderableType()),
        "swap_interval": int(fmt.swapInterval()),
        "samples": int(fmt.samples()),
        "depth_buffer_size": int(fmt.depthBufferSize()),
        "stencil_buffer_size": int(fmt.stencilBufferSize()),
        "alpha_buffer_size": int(fmt.alphaBufferSize()),
    }


def _collect_process_environment() -> dict[str, object]:
    return {
        "python": sys.version.split()[0],
        "platform": platform.platform(),
        "machine": platform.machine(),
        "env": {key: os.environ.get(key) for key in _INTERESTING_ENV_VARS},
    }


def _log_startup_environment(configured_logger, QtCore, QtGui, app) -> None:
    snapshot = _collect_process_environment()
    configured_logger.info(
        "topovec view environment platform=%s machine=%s python=%s env=%s",
        snapshot["platform"],
        snapshot["machine"],
        snapshot["python"],
        snapshot["env"],
    )
    screens = []
    for screen in app.screens():
        geometry = screen.geometry()
        screens.append(
            {
                "name": screen.name(),
                "size": (int(geometry.width()), int(geometry.height())),
                "dpr": float(screen.devicePixelRatio()),
            }
        )
    configured_logger.info(
        "Qt runtime platform_name=%s pyside=%s qt=%s default_surface_format=%s screens=%s",
        app.platformName(),
        getattr(QtCore, "__version__", "unknown"),
        QtCore.QLibraryInfo.version().toString(),
        _surface_format_snapshot(QtGui.QSurfaceFormat.defaultFormat()),
        screens,
    )
    if app.platformName() == "wayland":
        configured_logger.warning(
            "topovec view is known to exhibit OpenGL compositing artifacts on Qt Wayland in some Linux setups. "
            "If you see overbright or unstable rendering, rerun with QT_QPA_PLATFORM=xcb."
        )


def run_view(args: argparse.Namespace, *, argv: list[str] | None = None) -> int:
    configured_logger = configLog(logging.DEBUG if args.debug else logging.INFO)
    configured_logger.info(
        "Starting topovec view debug=%s startup_paths=%s max_live_scenes=%s",
        bool(args.debug),
        list(args.paths),
        int(args.max_live_scenes),
    )
    faulthandler.enable(file=sys.stderr, all_threads=True)
    previous_excepthook = _install_console_excepthook()
    previous_qt_message_handler = None
    try:
        QtCore, QtGui, QtWidgets = _load_qt()
        previous_qt_message_handler = _install_qt_message_logging(QtCore)
        from .viewport import QtViewport
        from .main_window import TvViewWindow

        app = QtWidgets.QApplication.instance()
        owns_app = app is None
        if app is None:
            configured_logger.debug("Creating QApplication for topovec view")
            QtGui.QSurfaceFormat.setDefaultFormat(QtViewport.create_surface_format())
            QtWidgets.QApplication.setApplicationName("topovec view")
            QtWidgets.QApplication.setOrganizationName("topovec")
            QtWidgets.QApplication.setApplicationDisplayName("topovec view")
            QtWidgets.QApplication.setAttribute(
                QtCore.Qt.ApplicationAttribute.AA_ShareOpenGLContexts
            )
            QtWidgets.QApplication.setAttribute(
                QtCore.Qt.ApplicationAttribute.AA_UseDesktopOpenGL
            )
            QtWidgets.QApplication.setHighDpiScaleFactorRoundingPolicy(
                QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough
            )
            app = QtWidgets.QApplication([sys.argv[0], *(argv or [])])
        else:
            configured_logger.debug("Reusing existing QApplication for topovec view")
        _log_startup_environment(configured_logger, QtCore, QtGui, app)

        configured_logger.debug("Constructing TvViewWindow")
        window = TvViewWindow(max_live_scenes=int(args.max_live_scenes))
        configured_logger.debug("TvViewWindow constructed")
        window.show()
        configured_logger.debug("TvViewWindow shown")
        if args.paths:
            window.load_paths(list(args.paths))
        if owns_app:
            configured_logger.debug("Entering topovec view event loop")
            return int(app.exec())
        return 0
    except Exception:
        configured_logger.exception("topovec view startup failed")
        return 1
    finally:
        if previous_qt_message_handler is not None:
            QtCore.qInstallMessageHandler(previous_qt_message_handler)
        sys.excepthook = previous_excepthook


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    return run_view(args, argv=argv)


__all__ = ["add_view_arguments", "build_parser", "main", "run_view"]



"""Low-level ModernGL render units used by scenes.

Render objects own shaders, GL programs, uniforms, and vertex arrays. They do
not own scene orchestration, camera-slot policy, or backend interaction.
"""

import moderngl as mgl
import numpy as np

from ..core import System


def _type_module_name(value) -> str:
    return str(getattr(type(value), "__module__", ""))


def _is_moderngl_proxy(value) -> bool:
    module = _type_module_name(value)
    return (
        module == "moderngl"
        or module.startswith("moderngl.")
        or module == "mgl"
        or module.startswith("mgl.")
        or module == "_moderngl"
        or module.startswith("_moderngl.")
    )


def _is_moderngl_context_like(value) -> bool:
    if isinstance(value, mgl.Context):
        return True
    return _is_moderngl_proxy(value) and type(value).__name__ == "Context"


def _release_nested_resources(value, *, seen: set[int] | None = None) -> None:
    if value is None:
        return
    if seen is None:
        seen = set()
    object_id = id(value)
    if object_id in seen:
        return
    seen.add(object_id)

    if isinstance(value, (str, bytes, bytearray, bool, int, float, np.generic)):
        return
    if isinstance(value, np.ndarray):
        return
    if isinstance(value, dict):
        for nested in value.values():
            _release_nested_resources(nested, seen=seen)
        return
    if isinstance(value, (list, tuple, set, frozenset)):
        for nested in value:
            _release_nested_resources(nested, seen=seen)
        return
    if _is_moderngl_context_like(value):
        return

    release = getattr(value, "release", None)
    if callable(release):
        try:
            release()
        except Exception:
            pass
        return

    # ModernGL exposes proxy objects such as uniforms whose attributes point
    # back into the shared GL context. Treat them as opaque leaves unless they
    # expose an explicit owned-resource release method handled above.
    if _is_moderngl_proxy(value):
        return

    attributes = getattr(value, "__dict__", None)
    if not isinstance(attributes, dict):
        return
    for nested in attributes.values():
        _release_nested_resources(nested, seen=seen)


class Render:
    """Base class for one ModernGL draw unit."""

    VERTEX = """
                #version 330
                in vec3 in_spin;
                void main() {
                    gl_Position = vec4(in_spin, 1.0);
                }
            """
    GEOMETRY = None
    FRAGMENT = """
                #version 330
                out vec4 f_color;
                void main() {
                    f_color = vec4(1.0,1.0,1.0,1.0);
                }
            """

    def __init__(   
        self, 
        mglctx: mgl.Context = None
    ):
        """Create the GL program and bind discovered uniforms as attributes."""
        self._ctx = mglctx

        self._prog = self._ctx.program(
            vertex_shader=self.VERTEX,
            geometry_shader=self.GEOMETRY,
            fragment_shader=self.FRAGMENT,
        )

        self.populate_uniforms()

    def populate_uniforms(self):
        """Expose all program uniforms as direct Python attributes."""
        for u in self._prog:
            obj = self._prog[u]
            if isinstance(obj, mgl.Uniform):
                self.__dict__[u] = obj

    def render(self, **kwargs):
        """Render the unit.

        Subclasses are expected to override this with concrete draw calls.
        """
        self._ctx.clear(0.60, 0.60, 0.60, 1.0)

    def list_properties(self):
        """Return render-local properties exposed through the scene contract."""
        return []

    def _replace_owned_geometry(
        self,
        build_geometry,
        build_vao=None,
        *,
        mesh_attr: str = "mesh",
        vao_attr: str = "_vao",
    ) -> tuple[object, object]:
        """Atomically replace a live `(mesh, vao)` pair.

        The old geometry is kept alive until the new mesh and VAO both exist.
        If construction fails, any partially built new resources are released
        and the previous live geometry stays attached to the render.
        """
        previous_mesh = getattr(self, mesh_attr, None)
        previous_vao = getattr(self, vao_attr, None)
        next_mesh = None
        next_vao = None
        try:
            if build_vao is None:
                next_mesh, next_vao = build_geometry()
            else:
                next_mesh = build_geometry()
                next_vao = build_vao(next_mesh)
        except Exception:
            if next_vao is not None:
                _release_nested_resources(next_vao)
            if next_mesh is not None:
                _release_nested_resources(next_mesh)
            raise

        setattr(self, mesh_attr, next_mesh)
        setattr(self, vao_attr, next_vao)
        if previous_vao is not None:
            _release_nested_resources(previous_vao)
        if previous_mesh is not None:
            _release_nested_resources(previous_mesh)
        return next_mesh, next_vao

    def release(self) -> None:
        """Release GL objects owned directly or indirectly by this render."""
        vao = getattr(self, "_vao", None)
        if vao is not None:
            try:
                vao.release()
            except Exception:
                pass
            self._vao = None

        prog = getattr(self, "_prog", None)
        if prog is not None:
            try:
                prog.release()
            except Exception:
                pass
            self._prog = None

        for name, value in list(getattr(self, "__dict__", {}).items()):
            if name in {"_ctx", "_prog", "_vao"}:
                continue
            if name.startswith("u_"):
                self.__dict__[name] = None
                continue
            if isinstance(value, np.ndarray):
                self.__dict__[name] = None
                continue
            _release_nested_resources(value)


####################################################################################


class SystemRender(Render):
    """Render that depends on a :class:`topovec.core.System`."""

    def __init__(
        self, 
        mglctx: mgl.Context = None,
        system: System = None,
    ):
        super().__init__(mglctx=mglctx)
        self._system = system



####################################################################################

class StateRender(SystemRender):
    """Render that additionally depends on a GPU-backed state buffer."""
    def __init__(
        self,
        mglctx: mgl.Context = None,
        system: System = None,
        stateBuffer: np.ndarray = None,
    ):
        super().__init__(mglctx=mglctx, system=system)
        self._state = stateBuffer

        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self._state.vbo, "3f", "in_spin"),
            ],
        )



import moderngl as mgl
import numpy as np

from .render import *
from .buffer import Float3DBuffer, IntBuffer
from ..core import NumericProperty, ListProperty, BooleanProperty


######################################################################################################

# GLSL functions for lattice points computation. For future usage.
#
# uniform ivec4 u_size; // Number of spins per each asis+number of atoms per unit cell.
# uniform mat3 u_generators; // Lattice generators.
# uniform vec3 u_representatives[]; // Atoms in the the unit cell.
#
# ivec4 reshape(int id) {
#     ivec4 stride;
#     stride.t = 1;
#     stride.z = stride.t*u_size.t;
#     stride.y = stride.z*u_size.z;
#     stride.x = stride.t*u_size.y;
#     return (id / stride) % u_size;
# }
#
# pos get_coordinates() {
#     ivec4 idx = reshape(gl_VertexID);
#     return u_generators*idx.xyz + u_representatives[idx.t];
# }

# /// Random number generator
# // https://stackoverflow.com/questions/4200224/random-noise-functions-for-glsl
# // A single iteration of Bob Jenkins' One-At-A-Time hashing algorithm.
# uint hash( uint x ) {
#     x += ( x << 10u );
#     x ^= ( x >>  6u );
#     x += ( x <<  3u );
#     x ^= ( x >> 11u );
#     x += ( x << 15u );
#     return x;
# }

# // Compound versions of the hashing algorithm I whipped together.
# uint hash( uvec2 v ) { return hash( v.x ^ hash(v.y)                         ); }
# uint hash( uvec3 v ) { return hash( v.x ^ hash(v.y) ^ hash(v.z)             ); }
# uint hash( uvec4 v ) { return hash( v.x ^ hash(v.y) ^ hash(v.z) ^ hash(v.w) ); }

# // Construct a float with half-open range [0:1] using low 23 bits.
# // All zeroes yields 0.0, all ones yields the next smallest representable value below 1.0.
# float floatConstruct( uint m ) {
#     const uint ieeeMantissa = 0x007FFFFFu; // binary32 mantissa bitmask
#     const uint ieeeOne      = 0x3F800000u; // 1.0 in IEEE binary32

#     m &= ieeeMantissa;                     // Keep only mantissa bits (fractional part)
#     m |= ieeeOne;                          // Add fractional part to 1.0

#     float  f = uintBitsToFloat( m );       // Range [1:2]
#     return f - 1.0;                        // Range [0:1]
# }

# // Pseudo-random value in half-open range [0:1].
# float random( float x ) { return floatConstruct(hash(floatBitsToUint(x))); }
# float random( vec2  v ) { return floatConstruct(hash(floatBitsToUint(v))); }
# float random( vec3  v ) { return floatConstruct(hash(floatBitsToUint(v))); }
# float random( vec4  v ) { return floatConstruct(hash(floatBitsToUint(v))); }

# vec3 random_vec() {
#     float r = floatConstruct(uint(gl_PrimitiveIDIn));
#     float g = random(r);
#     float b = random(g);
#     return vec3(r,g,b);
# }

# /// END random number generator


class FlowRender(StateRender):
    VERTEX = """
            #version 330

            const float M_PI = 3.1415926535897932384626433832795;

            uniform vec3 u_camera; // camera position in world coordinates
            uniform float u_color_shift;
            uniform bool u_neglect_z;
            uniform bool u_negate;

            in vec3 in_center; // center of atom
            in vec3 in_spin; // spin direction

            out VS_OUT {
                vec3 origin; // atom position in world coordiantes
                vec3 spin; // spin direction in spin space
                vec3 color;
            } vs_out;

            vec3 hsv2rgb(vec3 c) {
                vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
                vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
                return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
            }

            vec3 spin_to_color(vec3 spin) {
                float s = spin.z;
                float phi = (atan(spin.y, spin.x)/M_PI+1)/2;;
                vec3 hsv = vec3( phi, clamp(1-0.9*s, 0, 1), clamp(1+0.6*s, 0, 1) );
                hsv.x += u_color_shift;                        
                return hsv2rgb(hsv);
            }

            void main() {
                vec3 spin = normalize(in_spin);
                if (u_negate) spin = -spin;
                vs_out.origin = in_center;
                vs_out.spin = spin;
                if (u_neglect_z) {
                    vs_out.color = spin_to_color(vec3(spin.xy, 0));
                } else {
                    vs_out.color = spin_to_color(spin);
                };
            } 
        """

    GEOMETRY = """
            #version 330 core

            const float M_PI = 3.1415926535897932384626433832795;

            layout (lines_adjacency) in;

            layout (triangle_strip, max_vertices = 20) out;
//            layout (line_strip, max_vertices = 78) out;
//            layout (points, max_vertices = 4) out;

            uniform vec3 u_axis; // Zenith direction.
            uniform float u_level;
            uniform bool u_constant_theta;
            uniform int u_nparts=1;
            uniform int u_part_x=0;
            uniform int u_part_y=0;
            uniform int u_part_z=0;

            uniform mat4 u_view;
            uniform mat4 u_projection;

            uniform vec4 u_clip;

            uniform bool u_inverted;

            in VS_OUT {
                vec3 origin;
                vec3 spin;
                vec3 color;
            } vs_in[];

            out vec3 v_position;
            out vec3 v_color;
            out float v_clip;

            vec3 baryint(vec4 x, vec3 a, vec3 b, vec3 c, vec3 d) {
                return x[0]*a+x[1]*b+x[2]*c+x[3]*d;
            };

            void run0(vec4 a, vec4 b, vec2 phi) {
                // f(t)=f(0)+t*(f(1)-f(0))=0
                // t = f(0)/(f(0)-f(1)))
                float t = phi.x/(phi.x-phi.y);
                if (t<0 || t>1) return;
                v_position = mix(
                    baryint(a, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin), 
                    baryint(b, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin), 
                    t);
                v_color = mix(
                    baryint(a, vs_in[0].color, vs_in[1].color, vs_in[2].color, vs_in[3].color), 
                    baryint(b, vs_in[0].color, vs_in[1].color, vs_in[2].color, vs_in[3].color), 
                    t);

                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                v_clip = dot(vec4(v_position,1.0), u_clip); 
                EmitVertex();
            };

            void run(vec4 x, vec4 y, vec4 z, vec4 w, vec4 phi) {
                // There should be no point on the edge ab.
                run0(x, z, phi.xz);
                run0(y, z, phi.yz);
                run0(x, w, phi.xw);
                run0(y, w, phi.yw);
                run0(z, w, phi.zw); 

                /* For debug only.
                run0(x, z, phi.xz);
                run0(y, z, phi.yz);
                run0(x, w, phi.xw);
                run0(y, w, phi.yw);
                */

                EndPrimitive();
            };

            float potential(vec3 spin) {
                spin = normalize(spin); 
                if (u_constant_theta) {
                    return dot(spin, u_axis)-u_level;
                    // float theta = acos(dot(spin, u_axis));
                    // return cos(theta-0.5*M_PI*u_level);
                } else {
                    float phi = atan(spin.y, spin.x);
                    return cos(phi-M_PI*u_level);
                };
            };

            void process_simplex(vec4 x, vec4 y, vec4 z, vec4 w) {
                /* For debug only. Draw simplex in line-strip mode.
                v_color = vec3(0,0,0);
                v_position = baryint(x, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin);
                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
                v_position = baryint(y, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin);
                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
                v_position = baryint(z, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin);
                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
                v_position = baryint(w, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin);
                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
                v_position = baryint(x, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin);
                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
                v_position = baryint(z, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin);
                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
                v_position = baryint(y, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin);
                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
                v_position = baryint(w, vs_in[0].origin, vs_in[1].origin, vs_in[2].origin, vs_in[3].origin);
                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
                EndPrimitive();
                */

                vec4 phi = vec4(
                    potential(baryint(x, vs_in[0].spin, vs_in[1].spin, vs_in[2].spin, vs_in[3].spin)),
                    potential(baryint(y, vs_in[0].spin, vs_in[1].spin, vs_in[2].spin, vs_in[3].spin)),
                    potential(baryint(z, vs_in[0].spin, vs_in[1].spin, vs_in[2].spin, vs_in[3].spin)),
                    potential(baryint(w, vs_in[0].spin, vs_in[1].spin, vs_in[2].spin, vs_in[3].spin))
                );

                if (phi.x>0 && phi.y>0 && phi.z>0 && phi.w>0) return;
                if (phi.x<0 && phi.y<0 && phi.z<0 && phi.w<0) return;

                if      ((phi.x>0 && phi.y>0) || (phi.x<0 && phi.y<0)) run(x, y, z, w, phi.xyzw);
                else if ((phi.x>0 && phi.z>0) || (phi.x<0 && phi.z<0)) run(x, z, y, w, phi.xzyw);
                else run(z, y, x, w, phi.zyxw);
            };

            void main() {
                int w=u_nparts-1-u_part_x-u_part_y-u_part_z;
                vec4 o=vec4(u_part_x,u_part_y,u_part_z,w);
                
                process_simplex(
                    (o+vec4(1,0,0,0))/u_nparts,
                    (o+vec4(0,1,0,0))/u_nparts,
                    (o+vec4(0,0,1,0))/u_nparts,
                    (o+vec4(0,0,0,1))/u_nparts
                );
                
                if (u_part_x>0) {
                    process_simplex(
                        (o+vec4(0,0,0,1))/u_nparts,
                        (o+vec4(-1,0,1,1))/u_nparts,
                        (o+vec4(-1,1,1,0))/u_nparts,
                        (o+vec4(0,0,1,0))/u_nparts
                    );
                    
                    process_simplex(
                        (o+vec4(0,0,0,1))/u_nparts,
                        (o+vec4(0,1,0,0))/u_nparts,
                        (o+vec4(-1,1,1,0))/u_nparts,
                        (o+vec4(0,0,1,0))/u_nparts
                    );
                    
                    process_simplex(
                        (o+vec4(0,0,0,1))/u_nparts,
                        (o+vec4(-1,0,1,1))/u_nparts,
                        (o+vec4(-1,1,1,0))/u_nparts,
                        (o+vec4(-1,1,0,1))/u_nparts
                    );
                    
                    process_simplex(
                        (o+vec4(0,0,0,1))/u_nparts,
                        (o+vec4(0,1,0,0))/u_nparts,
                        (o+vec4(-1,1,1,0))/u_nparts,
                        (o+vec4(-1,1,0,1))/u_nparts
                    );
                };
            };
        """
    FRAGMENT = """
            #version 330

            uniform vec3 u_camera;
            uniform bool u_lighting;

            in vec3 v_position;
            in vec3 v_color;
            in float v_clip;

            out vec4 f_color;

            void main() {
                if(v_clip<0) discard;
                if (u_lighting) {
                    vec3 normal = normalize(cross(dFdx(v_position), dFdy(v_position)));
                    vec3 light_direction = -normalize(v_position-u_camera);
                    float inner_reflected = dot(light_direction, normal);
                    float intensity = 0.4+0.6*clamp(inner_reflected, 0.0, 1.0);
                    f_color = vec4(intensity*v_color.xyz, 1.0);
                    // f_color = vec4(abs(normal), 1.0); // DEBUG
                } else {
                    f_color = vec4(v_color.xyz, 1.0);
                };
            }

        """

    ANGLENAMES = ["polar angle", "azimuthal angle", "energy density"]

    def __init__(
        self,
        coordinatesBuffer: Float3DBuffer = None,
        maskBuffer=None,
        indexBuffer=None,
        scalarMaskBuffer=None,
        **kwargs,
    ):

        super().__init__(**kwargs)

        assert self._system.dim == 3

        self._coordinatesBuffer = coordinatesBuffer
        self._idx = indexBuffer

        self.init()

    def init(self):
        self._coordinatesBuffer.upload(self._system.spin_positions())

        self.init_vao()

        self.lighting = True
        self.negate = False
        self.color_shift = 0.81
        self._level = 0.0
        self.constant_theta = True
        self.axis = (0, 0, 1)
        self.nparts = 1
        self.neglect_z = False
        self._levels_number = 1.0
        self.cutx, self.cuty, self.cutz = -1.0, 0.0, 0.0
        self.cutphi, self.cuttheta = 0.0, 0.0
        self.levels = []

        self.update_levels()

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, value):
        self._level = value
        self.update_levels()

    @property
    def levels_number(self):
        return self._levels_number

    @levels_number.setter
    def levels_number(self, value):
        self._levels_number = value
        self.update_levels()

    def update_levels(self):
        self.levels = list(
            (self.level + n + 0.5) * 2.0 / self._levels_number - 1.0
            for n in range(int(self._levels_number))
        )

    def init_vao(self):
        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self._state.vbo, "3f/v", "in_spin"),
                (self._coordinatesBuffer.vbo, "3f/v", "in_center"),
            ],
            self._idx.ibo,
        )

    @property
    def angle_name(self):
        return self.ANGLENAMES[0 if self.constant_theta else 1]

    @angle_name.setter
    def angle_name(self, name):
        self.constant_theta = name == self.ANGLENAMES[0]
        # self.init_vao()

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        number_of_levels = int(self.levels_number)

        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())
        self.u_camera.value = tuple(camera_position)

        self.u_axis.value = self.axis
        self.u_constant_theta.value = self.constant_theta
        self.u_neglect_z.value = self.neglect_z  # or self.constant_theta
        self.u_color_shift.value = self.color_shift
        self.u_lighting.value = self.lighting
        self.u_negate.value = self.negate

        self._ctx.enable_only(mgl.DEPTH_TEST)

        nparts = int(self.nparts)
        assert nparts > 0
        self.u_nparts.value = nparts

        cutxyz = np.array([self.cutx, self.cuty, self.cutz])
        cutdir = np.array(
            [
                np.cos(np.pi * self.cutphi) * np.cos(np.pi * self.cuttheta),
                np.sin(np.pi * self.cutphi) * np.cos(np.pi * self.cuttheta),
                np.sin(np.pi * self.cuttheta),
            ]
        )
        self.u_clip.value = tuple(cutdir) + (-np.sum(cutdir * cutxyz),)

        for level in self.levels:
            # Render one surface level
            self.u_level.value = level
            for x in range(nparts):
                for y in range(nparts - x):
                    for z in range(nparts - x - y):
                        self.u_part_x.value = x
                        self.u_part_y.value = y
                        self.u_part_z.value = z
                        self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)

    def list_properties(self):
        maxxyz = self._system.center + self._system.lengths / 2
        minxyz = self._system.center - self._system.lengths / 2
        return [
            BooleanProperty(self, "lighting", title="Lighting"),
            BooleanProperty(self, "neglect_z", title="Ignore z-projection"),
            BooleanProperty(self, "negate", title="Negate spin"),
            ListProperty(self, "angle_name", title="Constant", values=self.ANGLENAMES),
            NumericProperty(
                self, "level", title="Level", min=-1.0, max=1.0, count=1000
            ),
            NumericProperty(
                self, "color_shift", title="Hue", min=0.0, max=1.0, count=360
            ),
            # NumericProperty(self, "nparts", title="Detalization", min=1.0, max=4.0, count=4),
            NumericProperty(
                self, "levels_number", title="#levels", min=1.0, max=10.0, count=10
            ),
            NumericProperty(
                self, "cutx", title="Cut X", min=minxyz[0], max=maxxyz[0], count=1000
            ),
            NumericProperty(
                self, "cuty", title="Cut Y", min=minxyz[1], max=maxxyz[1], count=1000
            ),
            NumericProperty(
                self, "cutz", title="Cut Z", min=minxyz[2], max=maxxyz[2], count=1000
            ),
            NumericProperty(
                self, "cutphi", title="Cut phi", min=0.0, max=2.0, count=1000
            ),
            NumericProperty(
                self, "cuttheta", title="Cut theta", min=-0.5, max=0.5, count=1000
            ),
        ]


####################################################################################################################


class AstraRender(FlowRender):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.level = 0.7
        self.npetals = 2
        self.ntheta = 1
        self.theta_shift = 0
        self.phi_shift = 0

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())
        self.u_camera.value = tuple(camera_position)

        self.u_constant_theta.value = True
        self.u_color_shift.value = self.color_shift
        self.u_level.value = self.level
        self.u_lighting.value = self.lighting
        self.u_neglect_z.value = self.neglect_z

        # self._ctx.enable_only(mgl.DEPTH_TEST)

        thetas = np.linspace(0, np.pi / 2, int(np.round(self.ntheta)), endpoint=False)
        theta_offset = np.pi / 2 / len(thetas) * self.theta_shift
        for theta in thetas:
            z, xy = np.sin(theta + theta_offset), np.cos(theta + theta_offset)
            k = int(np.round(np.cos(theta) * self.npetals)) + 1
            phi_offset = 2 * np.pi / k * self.phi_shift
            for phi in np.linspace(0, 2 * np.pi, k, endpoint=False) + phi_offset:
                x, y = xy * np.cos(phi), xy * np.sin(phi)
                self.u_axis.value = (x, y, z)
                self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)
        for theta in -thetas[1:]:
            z, xy = np.sin(theta + theta_offset), np.cos(theta + theta_offset)
            k = int(np.round(np.cos(theta) * self.npetals)) + 1
            phi_offset = 2 * np.pi / k * self.phi_shift
            for phi in np.linspace(0, 2 * np.pi, k, endpoint=False) + phi_offset:
                x, y = xy * np.cos(phi), xy * np.sin(phi)
                self.u_axis.value = (x, y, z)
                self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)

    def list_properties(self):
        return [
            BooleanProperty(self, "lighting", title="Lighting"),
            BooleanProperty(self, "neglect_z", title="Ignore z-projection"),
            NumericProperty(
                self, "level", title="Thickness", min=1.0, max=0.5, count=1000
            ),
            NumericProperty(
                self,
                "phi_shift",
                title="Polar angle offset",
                min=-1.0,
                max=1.0,
                count=1000,
            ),
            NumericProperty(
                self,
                "theta_shift",
                title="Azimuthal angle offset",
                min=-1.0,
                max=1.0,
                count=1000,
            ),
            NumericProperty(
                self, "npetals", title="Polar angle substeps", min=1, max=10, count=10
            ),
            NumericProperty(
                self, "ntheta", title="Azimuthal angle substeps", min=1, max=5, count=5
            ),
            NumericProperty(
                self, "color_shift", title="Hue", min=0.0, max=1.0, count=360
            ),
        ]



"""Curvilinear coordinate systems.

A coordinate system is a square local chart ``chi: U -> V`` on an ambient
space ``R^N``. Evaluating the chart at physical points ``x`` returns
``u = chi(x)``, the covariant tangent basis ``g_i = d x / d u_i``, and a
boolean mask marking the points where the chart is defined.

The numerical contract is:

- ``coords[..., i] = u_i(x)``
- ``basis[..., i, j] = d x_j / d u_i``
- ``valid[...]`` is ``True`` exactly on the domain of definition of the chart

The current implementation only supports square charts ``R^N -> R^N``.
Values returned outside ``valid == True`` are shape-correct but otherwise
unspecified; consumers are expected to honor ``valid`` explicitly.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable

import numpy as np

__all__ = [
    "CoordinateSampleND",
    "CoordinateSystemND",
    "CoordinateSystem2D",
    "CoordinateSystem3D",
    "EuclideanCoordinates",
    "AffineCoordinates",
    "CoordinateRemap",
    "RestrictedCoordinates",
    "CylindricalCoordinates",
    "PolarCoordinates",
    "AxisymmetricCoordinates",
]


def _validate_points(points, ndim: int) -> np.ndarray:
    """Validate that sampled ambient points have trailing dimension ``ndim``."""

    points = np.asarray(points)
    if points.shape[-1] != ndim:
        raise ValueError(f"Expected points with trailing dimension {ndim}, got {points.shape}.")
    return points


def _validate_square_matrix(matrix, ndim: int, name: str) -> np.ndarray:
    """Validate a square ``ndim x ndim`` matrix parameter."""

    matrix = np.asarray(matrix, dtype=np.float64)
    if matrix.shape != (ndim, ndim):
        raise ValueError(f"`{name}` must have shape ({ndim}, {ndim}), got {matrix.shape}.")
    return matrix


def _as_offset_vector(offset, ndim: int, name: str = "offset") -> np.ndarray:
    """Validate an affine offset vector in ``R^ndim``."""

    offset = np.asarray(offset, dtype=np.float64)
    if offset.shape != (ndim,):
        raise ValueError(f"`{name}` must have shape ({ndim},), got {offset.shape}.")
    return offset


def _identity_basis(shape: tuple[int, ...], ndim: int, dtype) -> np.ndarray:
    """Broadcast the Cartesian identity basis over a batch shape."""

    return np.broadcast_to(np.eye(ndim, dtype=dtype), shape + (ndim, ndim))


def _apply_inverse_coordinate_jacobian(basis: np.ndarray, inverse_jacobian: np.ndarray) -> np.ndarray:
    """Transform ``dx/du`` into ``dx/dv`` using the inverse coordinate Jacobian."""

    return np.einsum("...ia,...ij->...ja", basis, inverse_jacobian)


def _normalize_orientation(orientation: float) -> float:
    """Reduce an orientation sign to ``+1`` or ``-1``."""

    orientation = float(orientation)
    if orientation == 0:
        raise ValueError("`orientation` must be non-zero.")
    return 1.0 if orientation > 0 else -1.0


def _rotation_matrix_2d(angle: float) -> np.ndarray:
    """Return the standard 2D rotation matrix for ``angle``."""

    c = np.cos(angle)
    s = np.sin(angle)
    return np.array([[c, -s], [s, c]], dtype=np.float64)


def _rotation_matrix_3d(axis: int | np.ndarray | tuple[float, float, float], angle: float) -> np.ndarray:
    """Return the 3D rotation matrix around a Cartesian or explicit axis."""

    if np.isscalar(axis):
        axis_index = int(axis)
        if axis_index not in (0, 1, 2):
            raise ValueError("Integer `axis` must be one of 0, 1, 2.")
        axis_vec = np.zeros(3, dtype=np.float64)
        axis_vec[axis_index] = 1.0
    else:
        axis_vec = np.asarray(axis, dtype=np.float64)
        if axis_vec.shape != (3,):
            raise ValueError(f"`axis` must have shape (3,), got {axis_vec.shape}.")
        norm = np.linalg.norm(axis_vec)
        if norm == 0:
            raise ValueError("`axis` must be non-zero.")
        axis_vec = axis_vec / norm

    x, y, z = axis_vec
    c = np.cos(angle)
    s = np.sin(angle)
    cc = 1.0 - c
    return np.array(
        [
            [c + x * x * cc, x * y * cc - z * s, x * z * cc + y * s],
            [y * x * cc + z * s, c + y * y * cc, y * z * cc - x * s],
            [z * x * cc - y * s, z * y * cc + x * s, c + z * z * cc],
        ],
        dtype=np.float64,
    )


@dataclass(slots=True)
class CoordinateSampleND:
    """Evaluated square local chart.

    Parameters
    ----------
    coords:
        Array with shape ``(..., ndim)`` storing the curvilinear coordinates
        ``u = chi(x)``.
    basis:
        Array with shape ``(..., ndim, ndim)`` storing the covariant tangent
        basis vectors. By convention ``basis[..., i, :] = d x / d u_i``.
    valid:
        Boolean array with shape ``(...)`` marking where the chart is defined.

    Notes
    -----
    This object is purely structural. It does not attempt to sanitize values
    outside ``valid == True``; such values remain unspecified by contract.
    """

    coords: np.ndarray
    basis: np.ndarray
    valid: np.ndarray

    def __post_init__(self):
        self.coords = np.asarray(self.coords)
        self.basis = np.asarray(self.basis)
        self.valid = np.asarray(self.valid, dtype=bool)

        if self.coords.ndim < 1:
            raise ValueError("`coords` must have at least one dimension.")
        ndim = self.coords.shape[-1]
        if ndim <= 0:
            raise ValueError("The trailing coordinate dimension must be positive.")
        expected_basis_shape = self.coords.shape[:-1] + (ndim, ndim)
        if self.basis.shape != expected_basis_shape:
            raise ValueError(
                f"`basis` must have shape {expected_basis_shape}, got {self.basis.shape}."
            )
        if self.valid.shape != self.coords.shape[:-1]:
            raise ValueError(
                f"`valid` must have shape {self.coords.shape[:-1]}, got {self.valid.shape}."
            )

    @property
    def ndim(self) -> int:
        """Coordinate dimension ``N`` of the underlying square chart."""

        return int(self.coords.shape[-1])


class CoordinateSystemND(ABC):
    """Abstract square local chart on ``R^N``.

    A concrete coordinate system represents a local chart ``chi: U -> V`` on a
    square ambient space ``R^N``. Evaluating the chart at physical points ``x``
    returns:

    - coordinates ``u = chi(x)``
    - the covariant tangent basis ``g_i = d x / d u_i``
    - a boolean mask describing the domain of definition ``x in U``

    Implementations must respect the numerical contract:

    - input points have shape ``(..., ndim)``
    - ``coords`` has shape ``(..., ndim)``
    - ``basis`` has shape ``(..., ndim, ndim)``
    - ``basis[..., i, :] = d x / d u_i``
    - ``valid`` has shape ``(...)``

    The chart is allowed to be defined only on a patch. Mixed arrays of valid
    and invalid points are supported by construction through the ``valid`` mask.
    """

    _fixed_ndim: int | None = None

    def __init__(self, ndim: int | None = None):
        fixed_ndim = self._fixed_ndim
        if ndim is None:
            if fixed_ndim is None:
                raise TypeError("`ndim` must be provided for generic coordinate systems.")
            ndim = fixed_ndim
        ndim = int(ndim)
        if fixed_ndim is not None and ndim != fixed_ndim:
            raise ValueError(f"{self.__class__.__name__} requires ndim={fixed_ndim}, got {ndim}.")
        if ndim <= 0:
            raise ValueError("`ndim` must be positive.")
        self.ndim = ndim

    @abstractmethod
    def call(self, points: np.ndarray) -> CoordinateSampleND:
        """Evaluate the chart on physical points."""

        raise NotImplementedError

    def eval(self, points) -> CoordinateSampleND:
        """Validate points and return a checked :class:`CoordinateSampleND`."""

        points = _validate_points(points, self.ndim)
        sample = self.call(points)
        if not isinstance(sample, CoordinateSampleND):
            raise TypeError(
                f"{self.__class__.__name__}.call() must return CoordinateSampleND, got {type(sample).__name__}."
            )
        if sample.ndim != self.ndim:
            raise ValueError(
                f"{self.__class__.__name__} returned ndim={sample.ndim}, expected {self.ndim}."
            )
        expected_coords = points.shape
        expected_basis = points.shape[:-1] + (self.ndim, self.ndim)
        expected_valid = points.shape[:-1]
        if sample.coords.shape != expected_coords:
            raise ValueError(
                f"{self.__class__.__name__} returned coords shape {sample.coords.shape}, expected {expected_coords}."
            )
        if sample.basis.shape != expected_basis:
            raise ValueError(
                f"{self.__class__.__name__} returned basis shape {sample.basis.shape}, expected {expected_basis}."
            )
        if sample.valid.shape != expected_valid:
            raise ValueError(
                f"{self.__class__.__name__} returned valid shape {sample.valid.shape}, expected {expected_valid}."
            )
        return sample

    def __call__(self, points) -> CoordinateSampleND:
        """Alias for :meth:`eval`."""

        return self.eval(points)

    def coords(self, points) -> np.ndarray:
        """Return only the coordinate values ``u = chi(x)``."""

        return self.eval(points).coords

    def basis(self, points) -> np.ndarray:
        """Return only the covariant tangent basis ``d x / d u_i``."""

        return self.eval(points).basis

    def valid(self, points) -> np.ndarray:
        """Return only the domain-of-definition mask."""

        return self.eval(points).valid

    def mask(self, fn):
        """Build a scalar mask evaluated in this coordinate system.

        The callable is invoked as ``fn(*coords)`` where ``coords[i]`` is the
        ``i``-th curvilinear coordinate component returned by this chart. This
        is a convenience wrapper over ``MaskND.from_function(..., coordinates=self)``.
        """

        from .base import MaskND

        return MaskND.from_function(fn, coordinates=self)

    def affine(self, matrix, offset=None) -> "AffineCoordinates":
        """Precompose the chart with an ambient affine map."""

        return AffineCoordinates(self, matrix=matrix, offset=offset)

    def shift(self, offset) -> "AffineCoordinates":
        """Precompose the chart with an ambient translation.

        This is a convenience wrapper over :meth:`affine` with the identity
        ambient matrix and the requested translation vector.
        """

        offset = _as_offset_vector(offset, self.ndim, name="offset")
        return self.affine(np.eye(self.ndim, dtype=np.float64), offset=offset)

    def remap(
        self,
        map_fn: Callable[[np.ndarray], np.ndarray],
        inverse_jacobian_fn: Callable[[np.ndarray, np.ndarray], np.ndarray],
        valid_fn: Callable[[np.ndarray, np.ndarray], np.ndarray] | None = None,
    ) -> "CoordinateRemap":
        """Postcompose the chart with a coordinate-space map."""

        return CoordinateRemap(
            self,
            map_fn=map_fn,
            inverse_jacobian_fn=inverse_jacobian_fn,
            valid_fn=valid_fn,
        )

    def mix(self, matrix, offset=None) -> "CoordinateRemap":
        """Postcompose the chart with an invertible affine coordinate mix.

        If the base chart returns ``u``, the new coordinates are
        ``v = M u + b``. This is a convenience wrapper over :meth:`remap`
        rather than a separate public class.
        """

        matrix = _validate_square_matrix(matrix, self.ndim, "matrix")
        inverse_matrix = np.linalg.inv(matrix)
        if offset is None:
            offset = np.zeros(self.ndim, dtype=np.float64)
        offset = _as_offset_vector(offset, self.ndim)

        def map_fn(u: np.ndarray) -> np.ndarray:
            return u @ matrix.T + offset

        def inverse_jacobian_fn(u: np.ndarray, v: np.ndarray) -> np.ndarray:
            return np.broadcast_to(inverse_matrix, u.shape[:-1] + (self.ndim, self.ndim))

        return self.remap(map_fn=map_fn, inverse_jacobian_fn=inverse_jacobian_fn)

    def restrict(
        self, predicate: Callable[[np.ndarray, CoordinateSampleND], np.ndarray]
    ) -> "RestrictedCoordinates":
        """Restrict the chart to a smaller patch of its domain."""

        return RestrictedCoordinates(self, predicate=predicate)

    def cylindrize(
        self,
        radius: int,
        angle: int,
        angle_min: float = 0.0,
        angle_max: float = 2.0 * np.pi,
        tol: float = 1e-32,
        orientation: float = 1.0,
    ) -> "CylindricalCoordinates":
        """Postcompose the current chart with a cylindrical coordinate remap.

        The selected coordinate components are reinterpreted as a radius-like
        and angle-like pair, while all remaining coordinates are passed through
        unchanged.
        """

        return CylindricalCoordinates(
            self,
            radius=radius,
            angle=angle,
            angle_min=angle_min,
            angle_max=angle_max,
            tol=tol,
            orientation=orientation,
        )

    def rotate(
        self,
        angle: float,
        axis: int | np.ndarray | tuple[float, float, float] = 2,
        origin=None,
    ) -> "AffineCoordinates":
        """Rotate the physical input space when ``ndim`` is 2 or 3.

        This method exists on the generic base for usability. It is defined
        only for ambient dimensions 2 and 3 and raises ``NotImplementedError``
        otherwise.
        """

        if self.ndim == 2:
            if origin is None:
                origin = (0.0, 0.0)
            origin = _as_offset_vector(origin, 2, name="origin")
            rotation = _rotation_matrix_2d(float(angle))
        elif self.ndim == 3:
            if origin is None:
                origin = (0.0, 0.0, 0.0)
            origin = _as_offset_vector(origin, 3, name="origin")
            rotation = _rotation_matrix_3d(axis, float(angle))
        else:
            raise NotImplementedError("Ambient rotation is only implemented for ndim=2 and ndim=3.")
        offset = origin - origin @ rotation.T
        return self.affine(rotation, offset=offset)

    def axisymmetrize(
        self,
        axis: int = 2,
        angle_min: float = 0.0,
        angle_max: float = 2.0 * np.pi,
        tol: float = 1e-32,
        orientation: float = 1.0,
    ) -> "AxisymmetricCoordinates":
        """Lift a 2D meridional chart to a 3D axisymmetric chart.

        This method is only defined for ``ndim = 2`` and raises
        ``NotImplementedError`` otherwise.
        """

        if self.ndim != 2:
            raise NotImplementedError("Axisymmetric lifting is only implemented for ndim=2.")
        return AxisymmetricCoordinates(
            self,
            axis=axis,
            angle_min=angle_min,
            angle_max=angle_max,
            tol=tol,
            orientation=orientation,
        )


class CoordinateSystem2D(CoordinateSystemND):
    """Typed specialization of :class:`CoordinateSystemND` for ``ndim = 2``.

    This class is intentionally lightweight. It exists for intrinsically 2D
    charts and for future type-level constraints in ansatz code, not to mirror
    every generic coordinate transformation with a dedicated ``*2D`` wrapper.
    """

    _fixed_ndim = 2


class CoordinateSystem3D(CoordinateSystemND):
    """Typed specialization of :class:`CoordinateSystemND` for ``ndim = 3``.

    This class is intentionally lightweight. It exists for intrinsically 3D
    charts and for future type-level constraints in ansatz code, not to mirror
    every generic coordinate transformation with a dedicated ``*3D`` wrapper.
    """

    _fixed_ndim = 3


class EuclideanCoordinates(CoordinateSystemND):
    """Identity chart on ``R^N``.

    The returned coordinates coincide with the ambient Euclidean coordinates,
    the basis is the constant Cartesian basis, and ``valid`` is identically
    ``True``.
    """

    def call(self, points: np.ndarray) -> CoordinateSampleND:
        """Return the identity chart evaluated on Euclidean points."""

        points = _validate_points(points, self.ndim)
        dtype = np.result_type(points, np.float64)
        coords = np.asarray(points, dtype=dtype)
        basis = _identity_basis(points.shape[:-1], self.ndim, dtype)
        valid = np.ones(points.shape[:-1], dtype=bool)
        return CoordinateSampleND(coords=coords, basis=basis, valid=valid)


class AffineCoordinates(CoordinateSystemND):
    """Precompose a chart with an invertible affine ambient transformation.

    If the base chart is ``chi(x)`` and the affine ambient map is
    ``y = A x + b``, then this wrapper defines the new chart

    ``chi'(y) = chi(A^{-1} (y - b))``.

    The covariant tangent basis transforms as

    ``g'_i(y) = A g_i(A^{-1}(y - b))``.
    """

    def __init__(self, system: CoordinateSystemND, matrix, offset=None):
        self.system = system
        self.matrix = _validate_square_matrix(matrix, system.ndim, "matrix")
        self.inverse_matrix = np.linalg.inv(self.matrix)
        if offset is None:
            offset = np.zeros(system.ndim, dtype=np.float64)
        self.offset = _as_offset_vector(offset, system.ndim)
        super().__init__(system.ndim)

    def call(self, points: np.ndarray) -> CoordinateSampleND:
        """Evaluate the wrapped chart after applying the inverse affine map."""

        points = _validate_points(points, self.ndim)
        source_points = (np.asarray(points, dtype=np.result_type(points, np.float64)) - self.offset) @ self.inverse_matrix.T
        sample = self.system.eval(source_points)
        basis = sample.basis @ self.matrix.T
        return CoordinateSampleND(coords=sample.coords, basis=basis, valid=sample.valid)


class CoordinateRemap(CoordinateSystemND):
    """Postcompose a chart with a coordinate-space map.

    If the base chart is ``u = chi(x)`` and the new coordinates are defined by
    ``v = F(u)``, then this wrapper returns ``v`` as the coordinate values.

    The caller must provide the inverse differential
    ``d u / d v``. The tangent basis is updated as

    ``d x / d v_j = sum_i (d x / d u_i) (d u_i / d v_j)``.

    Parameters
    ----------
    map_fn:
        Function taking ``u`` with shape ``(..., ndim)`` and returning
        ``v = F(u)`` with the same shape.
    inverse_jacobian_fn:
        Function returning ``d u / d v`` with shape ``(..., ndim, ndim)``.
    valid_fn:
        Optional extra validity mask in the new coordinates. The final validity
        mask is ``base.valid & valid_fn(u, v)``.
    """

    def __init__(
        self,
        system: CoordinateSystemND,
        map_fn: Callable[[np.ndarray], np.ndarray],
        inverse_jacobian_fn: Callable[[np.ndarray, np.ndarray], np.ndarray],
        valid_fn: Callable[[np.ndarray, np.ndarray], np.ndarray] | None = None,
    ):
        self.system = system
        self.map_fn = map_fn
        self.inverse_jacobian_fn = inverse_jacobian_fn
        self.valid_fn = valid_fn
        super().__init__(system.ndim)

    def call(self, points: np.ndarray) -> CoordinateSampleND:
        """Evaluate the wrapped chart and then remap its coordinate values."""

        sample = self.system.eval(points)
        mapped_coords = np.asarray(self.map_fn(sample.coords))
        if mapped_coords.shape != sample.coords.shape:
            raise ValueError(
                f"`map_fn` must return shape {sample.coords.shape}, got {mapped_coords.shape}."
            )
        inverse_jacobian = np.asarray(self.inverse_jacobian_fn(sample.coords, mapped_coords))
        expected_jacobian = sample.coords.shape[:-1] + (self.ndim, self.ndim)
        if inverse_jacobian.shape != expected_jacobian:
            raise ValueError(
                f"`inverse_jacobian_fn` must return shape {expected_jacobian}, got {inverse_jacobian.shape}."
            )
        valid = sample.valid
        if self.valid_fn is not None:
            valid_extra = np.asarray(self.valid_fn(sample.coords, mapped_coords), dtype=bool)
            if valid_extra.shape != sample.valid.shape:
                raise ValueError(
                    f"`valid_fn` must return shape {sample.valid.shape}, got {valid_extra.shape}."
                )
            valid = valid & valid_extra
        basis = _apply_inverse_coordinate_jacobian(sample.basis, inverse_jacobian)
        return CoordinateSampleND(coords=mapped_coords, basis=basis, valid=valid)


class RestrictedCoordinates(CoordinateSystemND):
    """Restrict an existing chart to an explicit atlas patch.

    The wrapped chart is evaluated unchanged, but the final validity mask is

    ``valid' = valid & predicate(x, sample)``.
    """

    def __init__(
        self,
        system: CoordinateSystemND,
        predicate: Callable[[np.ndarray, CoordinateSampleND], np.ndarray],
    ):
        self.system = system
        self.predicate = predicate
        super().__init__(system.ndim)

    def call(self, points: np.ndarray) -> CoordinateSampleND:
        """Evaluate the wrapped chart and apply the extra patch predicate."""

        sample = self.system.eval(points)
        valid_extra = np.asarray(self.predicate(np.asarray(points), sample), dtype=bool)
        if valid_extra.shape != sample.valid.shape:
            raise ValueError(
                f"`predicate` must return shape {sample.valid.shape}, got {valid_extra.shape}."
            )
        return CoordinateSampleND(coords=sample.coords, basis=sample.basis, valid=sample.valid & valid_extra)


class CylindricalCoordinates(CoordinateSystemND):
    """Coordinate-space cylindrical remap of an existing chart.

    Let the wrapped chart return coordinates ``u`` and choose two distinct
    coordinate components ``u[radius]`` and ``u[angle]``. This wrapper defines
    new coordinates ``v`` by

    ``v[radius] = sqrt(u[radius]^2 + u[angle]^2)``

    ``v[angle] = angle_min + (angle_max - angle_min) * phi / (2π)``

    where ``phi = mod(orientation * atan2(u[angle], u[radius]), 2π)``. All
    remaining coordinate components are copied unchanged.

    The covariant tangent basis is transformed by the inverse differential
    ``d u / d v`` of the cylindrical remap, so ``basis[..., i, :]`` continues
    to store ``d x / d v_i`` in the ambient space of the wrapped chart.
    """

    def __init__(
        self,
        system: CoordinateSystemND,
        radius: int,
        angle: int,
        angle_min: float = 0.0,
        angle_max: float = 2.0 * np.pi,
        tol: float = 1e-32,
        orientation: float = 1.0,
    ):
        self.system = system
        self.radius = int(radius)
        self.angle = int(angle)
        if self.radius == self.angle:
            raise ValueError("`radius` and `angle` must be different indices.")
        if self.radius < 0 or self.radius >= system.ndim:
            raise ValueError(f"`radius` must be in [0, {system.ndim}), got {self.radius}.")
        if self.angle < 0 or self.angle >= system.ndim:
            raise ValueError(f"`angle` must be in [0, {system.ndim}), got {self.angle}.")
        self.angle_min = float(angle_min)
        self.angle_max = float(angle_max)
        if not self.angle_max > self.angle_min:
            raise ValueError("`angle_max` must be strictly greater than `angle_min`.")
        self.angle_period = self.angle_max - self.angle_min
        self.angle_scale = _normalize_orientation(orientation) * (2.0 * np.pi / self.angle_period)
        self.orientation = 1.0 if self.angle_scale > 0 else -1.0
        self.tol = float(tol)
        if self.tol < 0:
            raise ValueError("`tol` must be non-negative.")
        super().__init__(system.ndim)

    def call(self, points: np.ndarray) -> CoordinateSampleND:
        """Evaluate the cylindrical remap and its transformed tangent basis."""

        points = _validate_points(points, self.ndim)
        source_sample = self.system.eval(points)
        dtype = np.result_type(source_sample.coords, np.float64)
        source_coords = np.asarray(source_sample.coords, dtype=dtype)

        radial_x = np.asarray(source_coords[..., self.radius], dtype=dtype)
        radial_y = np.asarray(source_coords[..., self.angle], dtype=dtype)
        radius = np.sqrt(radial_x**2 + radial_y**2)
        psi = np.mod(np.arctan2(radial_y, radial_x), 2.0 * np.pi)
        angle_coord = self.angle_min + (self.angle_period / (2.0 * np.pi)) * np.mod(
            self.orientation * psi,
            2.0 * np.pi,
        )

        coords = np.array(source_coords, copy=True)
        coords[..., self.radius] = radius
        coords[..., self.angle] = angle_coord

        phi = self.angle_scale * (angle_coord - self.angle_min)
        cos_phi = np.cos(phi)
        sin_phi = np.sin(phi)

        inverse_jacobian = np.array(_identity_basis(points.shape[:-1], self.ndim, dtype), copy=True)
        inverse_jacobian[..., self.radius, self.radius] = cos_phi
        inverse_jacobian[..., self.angle, self.radius] = sin_phi
        inverse_jacobian[..., self.radius, self.angle] = -radius * sin_phi * self.angle_scale
        inverse_jacobian[..., self.angle, self.angle] = radius * cos_phi * self.angle_scale

        basis = _apply_inverse_coordinate_jacobian(source_sample.basis, inverse_jacobian)
        valid = source_sample.valid & (radius > self.tol)
        return CoordinateSampleND(coords=coords, basis=basis, valid=valid)


class PolarCoordinates(CylindricalCoordinates, CoordinateSystem2D):
    """Convenience specialization of :class:`CylindricalCoordinates` on ``R^2``.

    This class keeps the ergonomic public name ``PolarCoordinates`` but does
    not implement separate mathematics. It is exactly
    ``CylindricalCoordinates(EuclideanCoordinates(ndim=2), radius=0, angle=1, ...)``.
    """

    def __init__(
        self,
        angle_min: float = 0.0,
        angle_max: float = 2.0 * np.pi,
        tol: float = 1e-32,
        orientation: float = 1.0,
    ):
        super().__init__(
            EuclideanCoordinates(ndim=2),
            radius=0,
            angle=1,
            angle_min=angle_min,
            angle_max=angle_max,
            tol=tol,
            orientation=orientation,
        )


class AxisymmetricCoordinates(CoordinateSystem3D):
    """Lift a meridional 2D chart to a 3D axisymmetric chart.

    The wrapped meridional chart is evaluated on physical meridional points
    ``(rho, z)`` where ``rho`` is the Euclidean distance to the chosen rotation
    axis and ``z`` is the ambient coordinate along that axis.

    If the meridional chart returns coordinates ``(u_1, u_2)``, this class
    returns the 3D coordinates ``(u_1, u_2, a)``, where ``a`` is an angular
    source coordinate whose physical azimuth is

    ``phi(a) = orientation * 2π * (a - angle_min) / (angle_max - angle_min)``.

    The tangent basis is

    - ``d x / d u_k = (d rho / d u_k) e_r + (d z / d u_k) e_axis`` for
      ``k = 1, 2``
    - ``d x / d a = rho * d phi / d a * e_phi``

    The chart is singular on the rotation axis ``rho <= tol`` and those points
    are marked invalid via ``valid == False``.
    """

    def __init__(
        self,
        meridional: CoordinateSystemND,
        axis: int = 2,
        angle_min: float = 0.0,
        angle_max: float = 2.0 * np.pi,
        tol: float = 1e-32,
        orientation: float = 1.0,
    ):
        super().__init__()
        if meridional.ndim != 2:
            raise ValueError(
                f"`meridional` must have ndim=2, got {meridional.ndim} from {meridional.__class__.__name__}."
            )
        self.meridional = meridional
        self.axis = int(axis)
        if self.axis not in (0, 1, 2):
            raise ValueError("`axis` must be one of 0, 1, 2.")
        self.angle_min = float(angle_min)
        self.angle_max = float(angle_max)
        if not self.angle_max > self.angle_min:
            raise ValueError("`angle_max` must be strictly greater than `angle_min`.")
        self.angle_period = self.angle_max - self.angle_min
        self.angle_scale = _normalize_orientation(orientation) * (2.0 * np.pi / self.angle_period)
        self.orientation = 1.0 if self.angle_scale > 0 else -1.0
        self.tol = float(tol)
        if self.tol < 0:
            raise ValueError("`tol` must be non-negative.")
        self.transverse_axes = tuple(index for index in range(3) if index != self.axis)

    def call(self, points: np.ndarray) -> CoordinateSampleND:
        """Evaluate the lifted 3D axisymmetric chart on Euclidean points."""

        points = _validate_points(points, self.ndim)
        dtype = np.result_type(points, np.float64)
        transverse_x = np.asarray(points[..., self.transverse_axes[0]], dtype=dtype)
        transverse_y = np.asarray(points[..., self.transverse_axes[1]], dtype=dtype)
        axial = np.asarray(points[..., self.axis], dtype=dtype)

        rho = np.sqrt(transverse_x**2 + transverse_y**2)
        psi = np.mod(np.arctan2(transverse_y, transverse_x), 2.0 * np.pi)
        angle_coord = self.angle_min + (self.angle_period / (2.0 * np.pi)) * np.mod(
            self.orientation * psi,
            2.0 * np.pi,
        )
        meridional_points = np.stack([rho, axial], axis=-1)
        meridional_sample = self.meridional.eval(meridional_points)
        phi = self.angle_scale * (angle_coord - self.angle_min)

        er = np.zeros(points.shape, dtype=dtype)
        er[..., self.transverse_axes[0]] = np.cos(phi)
        er[..., self.transverse_axes[1]] = np.sin(phi)

        ephi = np.zeros(points.shape, dtype=dtype)
        ephi[..., self.transverse_axes[0]] = -np.sin(phi)
        ephi[..., self.transverse_axes[1]] = np.cos(phi)

        eaxis = np.zeros(points.shape, dtype=dtype)
        eaxis[..., self.axis] = 1.0

        basis = np.zeros(points.shape[:-1] + (3, 3), dtype=dtype)
        basis[..., 0, :] = (
            meridional_sample.basis[..., 0, 0, None] * er
            + meridional_sample.basis[..., 0, 1, None] * eaxis
        )
        basis[..., 1, :] = (
            meridional_sample.basis[..., 1, 0, None] * er
            + meridional_sample.basis[..., 1, 1, None] * eaxis
        )
        basis[..., 2, :] = (rho * self.angle_scale)[..., None] * ephi

        coords = np.concatenate([meridional_sample.coords, angle_coord[..., None]], axis=-1)
        valid = meridional_sample.valid & (rho > self.tol)
        return CoordinateSampleND(coords=coords, basis=basis, valid=valid)



import numpy as np
from numpy.testing import assert_allclose

import topovec.ansatz.frame as frame_module
from topovec.ansatz import (
    AffineCoordinates,
    AxisymmetricCoordinates,
    CoordinateRemap,
    CoordinateSampleND,
    CylindricalCoordinates,
    EuclideanCoordinates,
    PolarCoordinates,
)


def finite_coordinate_jacobian(system, points, step=1e-6):
    points = np.asarray(points, dtype=np.float64)
    jacobian = np.zeros(points.shape[:-1] + (system.ndim, system.ndim), dtype=np.float64)
    for axis in range(system.ndim):
        delta = np.zeros(system.ndim, dtype=np.float64)
        delta[axis] = step
        plus = system.coords(points + delta)
        minus = system.coords(points - delta)
        jacobian[..., :, axis] = (plus - minus) / (2.0 * step)
    return jacobian


def assert_basis_is_inverse_of_coordinate_jacobian(system, points, atol=1e-5, rtol=1e-5):
    sample = system(points)
    valid = sample.valid
    jacobian = finite_coordinate_jacobian(system, points)
    inverse_identity = np.einsum("...ia,...ja->...ij", jacobian, sample.basis)
    expected = np.eye(system.ndim)
    assert np.any(valid)
    assert_allclose(
        inverse_identity[valid],
        np.broadcast_to(expected, inverse_identity[valid].shape),
        atol=atol,
        rtol=rtol,
    )


def test_coordinate_sample_shapes_and_validation():
    sample = CoordinateSampleND(
        coords=np.zeros((4, 3, 2)),
        basis=np.broadcast_to(np.eye(2), (4, 3, 2, 2)),
        valid=np.ones((4, 3), dtype=bool),
    )
    assert sample.ndim == 2
    assert sample.coords.shape == (4, 3, 2)
    assert sample.basis.shape == (4, 3, 2, 2)
    assert sample.valid.shape == (4, 3)


def test_euclidean_coordinates_shapes_and_basis():
    points2 = np.array([[[0.0, 1.0], [2.0, -1.0]]], dtype=np.float64)
    points3 = np.array([[[0.0, 1.0, 2.0], [2.0, -1.0, 0.5]]], dtype=np.float64)

    system2 = EuclideanCoordinates(ndim=2)
    system3 = EuclideanCoordinates(ndim=3)

    sample2 = system2(points2)
    sample3 = system3(points3)

    assert isinstance(system2.shift((1.0, 2.0)), AffineCoordinates)
    assert isinstance(system3.shift((1.0, 2.0, 3.0)), AffineCoordinates)

    assert_allclose(sample2.coords, points2)
    assert_allclose(sample3.coords, points3)
    assert_allclose(sample2.basis, np.broadcast_to(np.eye(2), sample2.basis.shape))
    assert_allclose(sample3.basis, np.broadcast_to(np.eye(3), sample3.basis.shape))
    assert np.all(sample2.valid)
    assert np.all(sample3.valid)


def test_shift_and_rotation_precompose_physical_space():
    points = np.array([[2.0, 3.0], [-1.0, 4.0]], dtype=np.float64)
    base = EuclideanCoordinates(ndim=2)

    shifted = base.shift((1.5, -2.0))
    rotated = base.rotate(angle=np.pi / 2)

    assert_allclose(shifted.coords(points), points - np.array([1.5, -2.0]))
    expected_rot = points @ np.array([[0.0, -1.0], [1.0, 0.0]])
    assert_allclose(rotated.coords(points), expected_rot, atol=1e-12)
    assert_basis_is_inverse_of_coordinate_jacobian(shifted, points)
    assert_basis_is_inverse_of_coordinate_jacobian(rotated, points)


def test_coordinate_system_mask_sugar_uses_curvilinear_coordinates():
    points2 = np.array([[2.0, 3.0], [-1.0, 4.0]], dtype=np.float64)
    shifted2 = EuclideanCoordinates(ndim=2).shift((1.5, -2.0))
    mask2 = shifted2.mask(lambda x, z: x - 2.0 * z)
    assert_allclose(mask2(points2), (points2[:, 0] - 1.5) - 2.0 * (points2[:, 1] + 2.0))

    points3 = np.array([[2.0, 3.0, 4.0], [-1.0, 4.0, 0.5]], dtype=np.float64)
    shifted3 = EuclideanCoordinates(ndim=3).shift((1.0, -2.0, 0.5))
    mask3 = shifted3.mask(lambda x, y, z: x + y - z)
    assert_allclose(mask3(points3), (points3[:, 0] - 1.0) + (points3[:, 1] + 2.0) - (points3[:, 2] - 0.5))


def test_linear_coordinate_mix_and_coordinate_remap():
    points = np.array([[1.0, 2.0], [3.0, -4.0]], dtype=np.float64)
    base = EuclideanCoordinates(ndim=2)

    matrix = np.array([[2.0, 1.0], [1.0, 1.0]], dtype=np.float64)
    mix = base.mix(matrix=matrix, offset=(0.5, -1.0))
    sample = mix(points)
    expected_coords = points @ matrix.T + np.array([0.5, -1.0])
    expected_basis = np.broadcast_to(np.linalg.inv(matrix).T, sample.basis.shape)
    assert_allclose(sample.coords, expected_coords)
    assert_allclose(sample.basis, expected_basis)
    assert_basis_is_inverse_of_coordinate_jacobian(mix, points)

    remapped = CoordinateRemap(
        base,
        map_fn=lambda u: np.stack([u[..., 0] + 2.0 * u[..., 1], u[..., 1]], axis=-1),
        inverse_jacobian_fn=lambda u, v: np.broadcast_to(
            np.array([[1.0, -2.0], [0.0, 1.0]], dtype=np.float64),
            u.shape[:-1] + (2, 2),
        ),
    )
    remapped_sample = remapped(points)
    assert_allclose(remapped_sample.coords, np.stack([points[:, 0] + 2.0 * points[:, 1], points[:, 1]], axis=-1))
    assert_allclose(
        remapped_sample.basis,
        np.broadcast_to(np.array([[1.0, 0.0], [-2.0, 1.0]], dtype=np.float64), remapped_sample.basis.shape),
    )
    assert_basis_is_inverse_of_coordinate_jacobian(remapped, points)


def test_cylindrical_coordinates_2d_contract_and_singularity():
    points = np.array([[3.0, 4.0], [0.0, 0.0]], dtype=np.float64)
    cylindrical = CylindricalCoordinates(
        EuclideanCoordinates(ndim=2),
        radius=0,
        angle=1,
        angle_min=0.0,
        angle_max=2.0 * np.pi,
        tol=1e-12,
    )
    sample = cylindrical(points)

    assert_allclose(sample.coords[0], np.array([5.0, np.arctan2(4.0, 3.0)]))
    assert_allclose(sample.basis[0, 0], np.array([3.0 / 5.0, 4.0 / 5.0]))
    assert_allclose(sample.basis[0, 1], np.array([-4.0, 3.0]))
    assert bool(sample.valid[0])
    assert not bool(sample.valid[1])
    assert np.all(np.isfinite(sample.coords))
    assert np.all(np.isfinite(sample.basis))
    assert_basis_is_inverse_of_coordinate_jacobian(cylindrical, points[:1])


def test_polar_coordinates_is_convenience_specialization():
    points = np.array([[3.0, 4.0], [5.0, -12.0]], dtype=np.float64)
    polar = PolarCoordinates(tol=1e-12)
    cylindrical = CylindricalCoordinates(
        EuclideanCoordinates(ndim=2),
        radius=0,
        angle=1,
        angle_min=0.0,
        angle_max=2.0 * np.pi,
        tol=1e-12,
    )
    polar_sample = polar(points)
    cylindrical_sample = cylindrical(points)

    assert_allclose(polar_sample.coords, cylindrical_sample.coords)
    assert_allclose(polar_sample.basis, cylindrical_sample.basis)
    assert np.array_equal(polar_sample.valid, cylindrical_sample.valid)


def test_cylindrical_coordinates_3d_only_transform_selected_plane():
    points = np.array([[3.0, 4.0, 7.0], [5.0, -12.0, -1.0]], dtype=np.float64)
    cylindrical = CylindricalCoordinates(
        EuclideanCoordinates(ndim=3),
        radius=0,
        angle=1,
        angle_min=-1.0,
        angle_max=3.0,
        tol=1e-12,
    )
    sample = cylindrical(points)

    radius = np.sqrt(points[:, 0] ** 2 + points[:, 1] ** 2)
    psi = np.mod(np.arctan2(points[:, 1], points[:, 0]), 2.0 * np.pi)
    expected_angle = -1.0 + 4.0 * psi / (2.0 * np.pi)
    assert_allclose(sample.coords[:, 0], radius)
    assert_allclose(sample.coords[:, 1], expected_angle)
    assert_allclose(sample.coords[:, 2], points[:, 2])
    assert_basis_is_inverse_of_coordinate_jacobian(cylindrical, points)


def test_axisymmetric_coordinates_lift_meridional_chart():
    points = np.array(
        [
            [3.0, 4.0, 2.0],
            [5.0, 12.0, -1.0],
            [0.0, 0.0, 7.0],
        ],
        dtype=np.float64,
    )
    system = EuclideanCoordinates(ndim=2).axisymmetrize(
        axis=2,
        angle_min=1.0,
        angle_max=5.0,
        tol=1e-12,
    )
    sample = system(points)

    rho = np.sqrt(points[:, 0] ** 2 + points[:, 1] ** 2)
    psi = np.mod(np.arctan2(points[:, 1], points[:, 0]), 2.0 * np.pi)
    expected_angle = 1.0 + 4.0 * psi / (2.0 * np.pi)
    expected_coords = np.stack([rho, points[:, 2], expected_angle], axis=-1)
    assert_allclose(sample.coords, expected_coords)
    assert_allclose(sample.basis[0, 0], np.array([3.0 / 5.0, 4.0 / 5.0, 0.0]))
    assert_allclose(sample.basis[0, 1], np.array([0.0, 0.0, 1.0]))
    assert_allclose(
        sample.basis[0, 2],
        (5.0 * (2.0 * np.pi / 4.0)) * np.array([-4.0 / 5.0, 3.0 / 5.0, 0.0]),
    )
    assert bool(sample.valid[0])
    assert not bool(sample.valid[2])
    assert_basis_is_inverse_of_coordinate_jacobian(system, points[:2])


def test_restricted_patch_and_valid_propagation():
    base = PolarCoordinates(tol=1e-12)
    patch = base.restrict(predicate=lambda points, sample: sample.coords[..., 1] <= np.pi)
    points = np.array([[1.0, 1.0], [1.0, -1.0], [0.0, 0.0]], dtype=np.float64)
    valid = patch.valid(points)
    assert valid.tolist() == [True, False, False]


def test_cone_like_coordinates_from_meridional_mix_smoke():
    alpha = 0.37
    meridional = EuclideanCoordinates(ndim=2).mix(
        matrix=np.array(
            [
                [np.sin(alpha), np.cos(alpha)],
                [np.cos(alpha), -np.sin(alpha)],
            ],
            dtype=np.float64,
        )
    )
    cone_like = meridional.axisymmetrize(axis=2, tol=1e-12)
    points = np.array([[3.0, 4.0, 2.0], [5.0, 12.0, -1.0]], dtype=np.float64)
    sample = cone_like(points)

    rho = np.sqrt(points[:, 0] ** 2 + points[:, 1] ** 2)
    expected_first = np.sin(alpha) * rho + np.cos(alpha) * points[:, 2]
    assert_allclose(sample.coords[:, 0], expected_first)
    assert np.all(sample.valid)
    assert np.all(np.isfinite(sample.basis))


def test_dimension_specific_methods_use_runtime_dimension_checks():
    system2 = EuclideanCoordinates(ndim=2)
    system3 = EuclideanCoordinates(ndim=3)
    system4 = EuclideanCoordinates(ndim=4)

    assert isinstance(system2.rotate(np.pi / 3), AffineCoordinates)
    assert isinstance(system3.rotate(np.pi / 3, axis=1), AffineCoordinates)
    assert isinstance(system2.axisymmetrize(axis=2), AxisymmetricCoordinates)

    try:
        system4.rotate(0.1)
    except NotImplementedError:
        pass
    else:
        raise AssertionError("rotate() should reject dimensions other than 2 and 3.")

    try:
        system3.axisymmetrize(axis=2)
    except NotImplementedError:
        pass
    else:
        raise AssertionError("axisymmetrize() should reject dimensions other than 2.")


def test_public_coordinate_docstrings_present():
    assert frame_module.__doc__ is not None and "local chart" in frame_module.__doc__
    for cls in (
        frame_module.CoordinateSystemND,
        frame_module.AffineCoordinates,
        frame_module.CoordinateRemap,
        frame_module.RestrictedCoordinates,
        frame_module.CylindricalCoordinates,
        frame_module.PolarCoordinates,
        frame_module.AxisymmetricCoordinates,
    ):
        assert cls.__doc__ is not None
        assert len(cls.__doc__.strip()) > 40



import os
import sys

ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SRC = os.path.join(ROOT, "src")
if SRC not in sys.path:
    sys.path.insert(0, SRC)

project = "TopoVec"
author = "Igor Lobanov"
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "numpydoc",
    "sphinx_autodoc_typehints",
]
autosummary_generate = True
autosummary_generate_overwrite = True
autosummary_generated = ["api/_generated"]
numpydoc_show_class_members = False
numpydoc_class_members_toctree = True
autodoc_typehints = "description"
templates_path = ["_templates"]
exclude_patterns = ["_build"]
exclude_patterns.append("api/_generated")

html_theme = "pydata_sphinx_theme"
html_title = "TopoVec Documentation"
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "/")
